from functools import wraps
import inspect
from copy import copy
from typing import (
    Any,
    AsyncIterator,
    ClassVar,
    Optional,
    Union,
    Callable,
    Iterable,
    Iterator,
    List,
    Tuple,
    Dict,
)

from dlt.common.reflection.inspect import isasyncgenfunction, isgeneratorfunction
from dlt.common.typing import (
    AnyFun,
    AnyType,
    TDataItems,
    is_subclass,
    resolve_single_annotation,
)
from dlt.common.utils import get_callable_name, uniq_id

from dlt.extract.exceptions import (
    CreatePipeException,
    InvalidStepFunctionArguments,
    InvalidTransformerGeneratorFunction,
    ParametrizedResourceUnbound,
    PipeNotBoundToData,
    UnclosablePipe,
)
from dlt.extract.items import (
    ResolvablePipeItem,
    SupportsPipe,
    TPipeStep,
    TPipedDataItems,
)
from dlt.extract.items_transform import ItemTransform
from dlt.extract.utils import (
    check_compat_transformer,
    simulate_func_call,
    wrap_compat_transformer,
    wrap_resource_gen,
    wrap_async_iterator,
)


class ForkPipe(ItemTransform[ResolvablePipeItem, Dict[str, Any]]):
    placement_affinity: ClassVar[float] = 2

    def __init__(self, pipe: "Pipe", step: int = -1, copy_on_fork: bool = False) -> None:
        """A transformer that forks the `pipe` and sends the data items to forks added via `add_pipe` method."""
        self._pipes: List[Tuple["Pipe", int]] = []
        self.copy_on_fork = copy_on_fork
        """If true, the data items going to a forked pipe will be copied"""
        self.add_pipe(pipe, step)

    def add_pipe(self, pipe: "Pipe", step: int = -1) -> None:
        if pipe not in self._pipes:
            self._pipes.append((pipe, step))

    def has_pipe(self, pipe: "Pipe") -> bool:
        return pipe in [p[0] for p in self._pipes]

    def __call__(self, item: TDataItems, meta: Any = None) -> Iterator[ResolvablePipeItem]:
        for i, (pipe, step) in enumerate(self._pipes):
            if i == 0 or not self.copy_on_fork:
                _it = item
            else:
                # shallow copy the item
                _it = copy(item)
            # always start at the beginning
            yield ResolvablePipeItem(_it, step, pipe, meta)


class Pipe(SupportsPipe):
    def __init__(self, name: str, steps: List[TPipeStep] = None, parent: "Pipe" = None) -> None:
        self.name = name
        self._gen_idx = 0
        self._steps: List[TPipeStep] = []
        self.parent = parent
        self.instance_id = uniq_id()
        # add the steps, this will check and mod transformations
        if steps:
            for index, step in enumerate(steps):
                self.insert_step(step, index)

    @classmethod
    def from_data(
        cls,
        name: str,
        gen: Union[Iterable[TPipedDataItems], Iterator[TPipedDataItems], AnyFun],
        parent: "Pipe" = None,
    ) -> "Pipe":
        return cls(name, [gen], parent=parent)

    @property
    def is_empty(self) -> bool:
        """Checks if pipe contains any steps"""
        return len(self._steps) == 0

    @property
    def has_parent(self) -> bool:
        return self.parent is not None

    @property
    def is_data_bound(self) -> bool:
        """Checks if pipe is bound to data and can be iterated. Pipe is bound if has a parent that is bound xor is not empty."""
        if self.has_parent:
            return self.parent.is_data_bound
        else:
            return not self.is_empty

    @property
    def gen(self) -> TPipeStep:
        """A data generating step"""
        return self._steps[self._gen_idx]

    @property
    def tail(self) -> TPipeStep:
        return self._steps[-1]

    @property
    def steps(self) -> List[TPipeStep]:
        return self._steps

    def find(self, *step_type: AnyType) -> int:
        """Finds a step with object of type `step_type`"""
        found = self.find_all(step_type)
        return found[0] if found else -1

    def find_all(self, *step_type: AnyType) -> List[int]:
        """Finds all steps with object of type `step_type`"""
        return [i for i, v in enumerate(self._steps) if isinstance(v, step_type)]

    def get_by_type(self, *step_type: AnyType) -> TPipeStep:
        """Gets first step found with object of type `step_type`"""
        return next((v for v in self._steps if isinstance(v, step_type)), None)

    def remove_by_type(self, *step_type: AnyType) -> int:
        """Deletes first step found with object of type `step_type`, returns previous index"""
        step_index = self.find(*step_type)
        if step_index >= 0:
            self.remove_step(step_index)
        return step_index

    def __getitem__(self, i: int) -> TPipeStep:
        return self._steps[i]

    def __len__(self) -> int:
        return len(self._steps)

    def fork(self, child_pipe: "Pipe", child_step: int = -1, copy_on_fork: bool = False) -> "Pipe":
        if len(self._steps) == 0:
            raise CreatePipeException(self.name, f"Cannot fork to empty pipe `{child_pipe}`")
        fork_step = self.tail
        if not isinstance(fork_step, ForkPipe):
            fork_step = ForkPipe(child_pipe, child_step, copy_on_fork)
            # always add this at the end
            self.insert_step(fork_step, len(self))
        else:
            if not fork_step.has_pipe(child_pipe):
                fork_step.add_pipe(child_pipe, child_step)
        return self

    def append_step(self, step: TPipeStep) -> "Pipe":
        """Appends pipeline step. On first added step performs additional verification if step is a valid data generator"""
        steps_count = len(self._steps)

        if steps_count == 0 and not self.has_parent:
            self._verify_head_step(step)
        else:
            step = self._wrap_transform_step_meta(steps_count, step)

        # find the insert position using particular
        if steps_count > 0:
            affinity = step.placement_affinity if isinstance(step, ItemTransform) else 0
            for index in reversed(range(0, steps_count)):
                step_at_idx = self._steps[index]
                affinity_at_idx = (
                    step_at_idx.placement_affinity if isinstance(step_at_idx, ItemTransform) else 0
                )
                if affinity_at_idx <= affinity:
                    self._insert_at_pos(step, index + 1)
                    return self
            # insert at the start due to strong affinity
            self._insert_at_pos(step, 0)
        else:
            self._steps.append(step)
        return self

    def insert_step(self, step: TPipeStep, index: int) -> "Pipe":
        """Inserts step at a given index in the pipeline. Allows prepending only for transformers"""
        steps_count = len(self._steps)
        if steps_count == 0:
            return self.append_step(step)
        if index == 0:
            if not self.has_parent:
                raise CreatePipeException(
                    self.name,
                    "You cannot insert a step before head of the resource that is not a"
                    " transformer",
                )
        step = self._wrap_transform_step_meta(index, step)
        self._insert_at_pos(step, index)
        return self

    def remove_step(self, index: int) -> None:
        """Removes steps at a given index. Gen step cannot be removed"""
        if index == self._gen_idx:
            raise CreatePipeException(
                self.name,
                f"Step at `{index=:}` holds a data generator for this pipe and cannot be removed",
            )
        self._steps.pop(index)
        if index < self._gen_idx:
            self._gen_idx -= 1

    def replace_gen(self, gen: TPipeStep) -> None:
        """Replaces data generating step. Assumes that you know what are you doing"""
        assert not self.is_empty
        self._steps[self._gen_idx] = gen

    def close(self) -> None:
        """Closes pipe generator"""
        gen = self.gen
        # NOTE: async generator are wrapped in generators
        if inspect.isgenerator(gen):
            gen.close()
        else:
            raise UnclosablePipe(self.name, gen)

    def full_pipe(self) -> "Pipe":
        """Creates a pipe that from the current and all the parent pipes."""
        # prevent creating full pipe with unbound heads
        if self.has_parent:
            self._ensure_transform_step(self._gen_idx, self.gen)
        else:
            self.ensure_gen_bound()

        if self.has_parent:
            steps = self.parent.full_pipe().steps
        else:
            steps = []

        steps.extend(self._steps)
        p = Pipe(self.name, [])
        # set the steps so they are not evaluated again
        p._steps = steps
        # return pipe with resolved dependencies
        return p

    def ensure_gen_bound(self) -> None:
        """Verifies that gen step is bound to data"""
        head = self.gen
        if not callable(head):
            return
        sig = inspect.signature(head)
        try:
            # must bind without arguments
            sig.bind()
        except TypeError as ex:
            callable_name = get_callable_name(head)
            raise ParametrizedResourceUnbound(
                self.name,
                callable_name,
                sig.replace(parameters=list(sig.parameters.values())[1:]),
                "resource",
                str(ex),
            )

    def evaluate_gen(self) -> None:
        """Lazily evaluate gen of the pipe when creating PipeIterator. Allows creating multiple use pipes from generator functions and lists"""
        if not self.is_data_bound:
            raise PipeNotBoundToData(self.name, self.has_parent)

        gen = self.gen
        if not self.has_parent:
            if callable(gen):
                try:
                    # must be parameter-less callable or parameters must have defaults
                    self.replace_gen(gen())  # type: ignore
                except TypeError as ex:
                    raise ParametrizedResourceUnbound(
                        self.name,
                        get_callable_name(gen),
                        inspect.signature(gen),
                        "resource",
                        str(ex),
                    )
            # otherwise it must be an iterator
            gen = self.gen
            if isinstance(gen, Iterable):
                self.replace_gen(iter(gen))
        else:
            # verify if transformer can be called
            self._ensure_transform_step(self._gen_idx, gen)

        # wrap async generator
        if isinstance(self.gen, AsyncIterator):
            self.replace_gen(wrap_async_iterator(self.gen))

        # evaluate transforms
        for step_no, step in enumerate(self._steps):
            # print(f"pipe {self.name} step no {step_no} step({step})")
            if isinstance(step, ItemTransform):
                self._steps[step_no] = step.bind(self)

    def bind_gen(self, *args: Any, **kwargs: Any) -> Any:
        """Finds and wraps with `args` + `kwargs` the callable generating step in the resource pipe and then replaces the pipe gen with the wrapped one"""
        # try:
        gen = self._wrap_gen(*args, **kwargs)
        self.replace_gen(gen)
        return gen

    def sim_gen(
        self, *args: Any, **kwargs: Any
    ) -> Tuple[inspect.Signature, inspect.Signature, inspect.BoundArguments]:
        """Simulates calling the gen element"""
        # skip the data item argument for transformers
        args_to_skip = 1 if self.has_parent else 0
        # simulate function call
        return simulate_func_call(self.gen, args_to_skip, *args, **kwargs)

    def _wrap_gen(self, *args: Any, **kwargs: Any) -> Any:
        """Finds and wraps with `args` + `kwargs` the callable generating step in the resource pipe."""
        head = self.gen
        _data: Any = None

        sig, _, _ = self.sim_gen(*args, **kwargs)
        assert callable(head)

        # create wrappers with partial
        if self.has_parent:
            _data = wrap_compat_transformer(self.name, head, sig, *args, **kwargs)
        else:
            if self._should_eval_on_bind(head, sig):
                # call immediately to replace gen, this will happen before pipe executes
                _data = head(*args, **kwargs)
            else:
                _data = wrap_resource_gen(self.name, head, sig, *args, **kwargs)
        return _data

    def _should_eval_on_bind(self, gen: Any, sig: inspect.Signature) -> bool:
        """Checks if gen should be evaluated when args are bound to it. Standard behavior is to
        eval during extraction. Currently only when resource is returned we evaluate
        """
        if not isasyncgenfunction(gen) and not isgeneratorfunction(gen):
            # TODO: some base methods of DltSource and DltResource should be moved to common
            #  below we import DltResource but Pipe class should not be dependent on it
            from dlt.extract.resource import DltResource

            if sig.return_annotation != inspect.Signature.empty:
                # globals will contain DltResource which we want to resolve
                return_annotation = resolve_single_annotation(
                    sig.return_annotation, globalns=globals()
                )
                return is_subclass(return_annotation, DltResource)

        return False

    def _verify_head_step(self, step: TPipeStep) -> None:
        # first element must be Iterable, Iterator or Callable in resource pipe
        if not isinstance(step, (Iterable, Iterator, AsyncIterator)) and not callable(step):
            raise CreatePipeException(
                self.name,
                "A head of a resource pipe must be: [Iterable, Iterator, AsyncIterator, Callable]",
            )

    def _wrap_transform_step_meta(self, step_no: int, step: TPipeStep) -> TPipeStep:
        # step must be a callable: a transformer or a transformation
        if isinstance(step, (Iterable, Iterator)) and not callable(step):
            if self.has_parent:
                raise CreatePipeException(
                    self.name, "Iterable or Iterator cannot be a step in transformer pipe"
                )
            else:
                raise CreatePipeException(
                    self.name, "Iterable or Iterator can only be a first step in resource pipe"
                )

        if not callable(step):
            raise CreatePipeException(
                self.name,
                "Pipe step must be a callable taking one data item as argument and optional second"
                " meta argument",
            )
        else:
            # check the signature
            sig = inspect.signature(step)
            meta_arg = check_compat_transformer(self.name, step, sig)
            if meta_arg is None:
                # add meta parameter when not present

                meta_arg = inspect.Parameter(
                    "meta", inspect._ParameterKind.KEYWORD_ONLY, default=None
                )
                kwargs_arg = next(
                    (p for p in sig.parameters.values() if p.kind == inspect.Parameter.VAR_KEYWORD),
                    None,
                )
                if kwargs_arg:
                    # pass meta in variadic
                    new_sig = sig
                else:
                    params = list(sig.parameters.values()) + [meta_arg]
                    new_sig = sig.replace(parameters=params)

                orig_step = step

                @wraps(step)
                def _partial(*args: Any, **kwargs: Any) -> Any:
                    # orig step does not have meta
                    kwargs.pop("meta", None)
                    # del kwargs["meta"]
                    return orig_step(*args, **kwargs)

                setattr(_partial, "__signature__", new_sig)
                step = _partial

            # verify the step callable, gen may be parametrized and will be evaluated at run time
            if not self.is_empty:
                self._ensure_transform_step(step_no, step)
        return step

    def _ensure_transform_step(self, step_no: int, step: TPipeStep) -> None:
        """Verifies that `step` is a valid callable to be a transform step of the pipeline"""
        assert callable(step), f"{step} must be callable"

        sig = inspect.signature(step)
        try:
            # get eventually modified sig
            sig.bind("item", meta="meta")
        except TypeError as ty_ex:
            callable_name = get_callable_name(step)
            if step_no == self._gen_idx:
                # error for gen step
                if len(sig.parameters) == 0:
                    raise InvalidTransformerGeneratorFunction(self.name, callable_name, sig, code=1)
                else:
                    # show the sig without first argument
                    raise ParametrizedResourceUnbound(
                        self.name,
                        callable_name,
                        sig.replace(parameters=list(sig.parameters.values())[1:]),
                        "transformer",
                        str(ty_ex),
                    )
            else:
                raise InvalidStepFunctionArguments(self.name, callable_name, sig, str(ty_ex))

    def _clone(self, new_name: str = None) -> "Pipe":
        """Clones the pipe steps, optionally renaming the pipe. Used internally to clone a list of connected pipes."""
        new_parent = self.parent
        p = Pipe(new_name or self.name, [], new_parent)
        # keep instance id
        p.instance_id = self.instance_id
        p._steps = self._steps.copy()
        return p

    def _insert_at_pos(self, step: Any, index: int) -> None:
        # shift right if no parent
        if index == 0 and not self.has_parent:
            # put after gen
            index += 1
        # actually insert in the list
        self._steps.insert(index, step)
        # increase the _gen_idx if added before generator
        if index <= self._gen_idx:
            self._gen_idx += 1

    def __repr__(self) -> str:
        if self.has_parent:
            bound_str = " data bound to " + repr(self.parent)
        else:
            bound_str = ""
        return f"Pipe {self.name} [steps: {len(self._steps)}] at {id(self)}{bound_str}"



import fnmatch
import hashlib
import os
import yaml
import posixpath
from pathlib import Path
from typing import Dict, NamedTuple, Sequence, Tuple, List, Literal
from dlt._workspace.cli.exceptions import VerifiedSourceRepoError

from dlt.common.libs import git
from dlt.common.storages import FileStorage
from dlt.common.typing import TypedDict

from dlt.common.reflection.utils import get_module_docstring

from dlt._workspace.cli import utils
from dlt._workspace.cli.requirements import SourceRequirements

TSourceType = Literal["core", "verified", "template"]

SOURCES_INIT_INFO_ENGINE_VERSION = 1

SOURCES_MODULE_NAME = "sources"
TEMPLATES_MODULE_NAME = "_templates"
CORE_SOURCE_TEMPLATE_MODULE_NAME = "_core_source_templates"
SINGLE_FILE_TEMPLATE_MODULE_NAME = "_single_file_templates"

SOURCES_INIT_INFO_FILE = ".sources"
IGNORE_FILES = ["*.py[cod]", "*$py.class", "__pycache__", "py.typed", "requirements.txt"]
IGNORE_VERIFIED_SOURCES = [".*", "_*"]
IGNORE_CORE_SOURCES = [
    ".*",
    "_*",
    "helpers",
    SINGLE_FILE_TEMPLATE_MODULE_NAME,
    CORE_SOURCE_TEMPLATE_MODULE_NAME,
]
PIPELINE_FILE_SUFFIX = "_pipeline.py"
# hardcode default template files here
TEMPLATE_FILES = [".gitignore", ".dlt/config.toml"]
DEFAULT_PIPELINE_TEMPLATE = "default_pipeline.py"


class SourceConfiguration(NamedTuple):
    source_type: TSourceType
    source_module_prefix: str
    storage: FileStorage
    src_pipeline_script: str
    dest_pipeline_script: str
    files: List[str]
    requirements: SourceRequirements
    doc: str
    is_default_template: bool


class TVerifiedSourceFileEntry(TypedDict):
    commit_sha: str
    git_sha: str
    sha3_256: str


class TVerifiedSourceFileIndex(TypedDict):
    is_dirty: bool
    last_commit_sha: str
    last_commit_timestamp: str
    files: Dict[str, TVerifiedSourceFileEntry]
    dlt_version_constraint: str


class TVerifiedSourcesFileIndex(TypedDict):
    engine_version: int
    sources: Dict[str, TVerifiedSourceFileIndex]


def _save_dot_sources(index: TVerifiedSourcesFileIndex) -> None:
    with open(utils.make_dlt_settings_path(SOURCES_INIT_INFO_FILE), "w", encoding="utf-8") as f:
        yaml.dump(index, f, allow_unicode=True, default_flow_style=False, sort_keys=False)


def _load_dot_sources() -> TVerifiedSourcesFileIndex:
    try:
        with open(utils.make_dlt_settings_path(SOURCES_INIT_INFO_FILE), "r", encoding="utf-8") as f:
            index: TVerifiedSourcesFileIndex = yaml.safe_load(f)
            if not index:
                raise FileNotFoundError(SOURCES_INIT_INFO_FILE)
            return index
    except FileNotFoundError:
        return {"engine_version": SOURCES_INIT_INFO_ENGINE_VERSION, "sources": {}}


def _merge_remote_index(
    local_index: TVerifiedSourceFileIndex,
    remote_index: TVerifiedSourceFileIndex,
    remote_modified: Dict[str, TVerifiedSourceFileEntry],
    remote_deleted: Dict[str, TVerifiedSourceFileEntry],
) -> TVerifiedSourceFileIndex:
    # update all modified files
    local_index["files"].update(remote_modified)
    # delete all deleted
    for deleted in remote_deleted:
        del local_index["files"][deleted]
    # update global info
    local_index["is_dirty"] = remote_index["is_dirty"]
    local_index["last_commit_sha"] = remote_index["last_commit_sha"]
    local_index["last_commit_timestamp"] = remote_index["last_commit_timestamp"]
    local_index["dlt_version_constraint"] = remote_index["dlt_version_constraint"]

    return local_index


def load_verified_sources_local_index(source_name: str) -> TVerifiedSourceFileIndex:
    return _load_dot_sources()["sources"].get(
        source_name,
        {
            "is_dirty": False,
            "last_commit_sha": None,
            "last_commit_timestamp": None,
            "files": {},
            "dlt_version_constraint": ">=0.1.0",
        },
    )


def save_verified_source_local_index(
    source_name: str,
    remote_index: TVerifiedSourceFileIndex,
    remote_modified: Dict[str, TVerifiedSourceFileEntry],
    remote_deleted: Dict[str, TVerifiedSourceFileEntry],
) -> None:
    all_sources = _load_dot_sources()
    local_index = all_sources["sources"].setdefault(source_name, remote_index)
    _merge_remote_index(local_index, remote_index, remote_modified, remote_deleted)
    _save_dot_sources(all_sources)


def get_remote_source_index(
    repo_path: str, files: Sequence[str], dlt_version_constraint: str
) -> TVerifiedSourceFileIndex:
    with git.get_repo(repo_path) as repo:
        tree = repo.tree()
        commit_sha = repo.head.commit.hexsha
        files_sha: Dict[str, TVerifiedSourceFileEntry] = {}
        for file in files:
            posix_file = os.path.join(repo_path, file)
            posix_file = os.path.relpath(posix_file, repo.working_dir)
            posix_file = posixpath.join(*Path(posix_file).parts)
            try:
                blob_sha3 = tree.join(posix_file).hexsha
            except KeyError:
                # if directory is dirty and we do not have git sha
                blob_sha3 = None

            with open(os.path.join(repo_path, file), "rb") as f:
                file_blob = f.read()
            files_sha[file] = {
                "commit_sha": commit_sha,
                "git_sha": blob_sha3,
                "sha3_256": hashlib.sha3_256(file_blob).hexdigest(),
            }

        return {
            "is_dirty": git.is_dirty(repo),
            "last_commit_sha": commit_sha,
            "last_commit_timestamp": repo.head.commit.committed_datetime.isoformat(),
            "files": files_sha,
            "dlt_version_constraint": dlt_version_constraint,
        }


def get_sources_names(sources_storage: FileStorage, source_type: TSourceType) -> List[str]:
    candidates: List[str] = []

    if source_type == "template":
        for name in sources_storage.list_folder_files(".", to_root=False):
            if name.endswith(PIPELINE_FILE_SUFFIX):
                candidates.append(name.replace(PIPELINE_FILE_SUFFIX, ""))
    else:
        ignore_cases = IGNORE_VERIFIED_SOURCES if source_type == "verified" else IGNORE_CORE_SOURCES
        for name in [
            n
            for n in sources_storage.list_folder_dirs(".", to_root=False)
            if not any(fnmatch.fnmatch(n, ignore) for ignore in ignore_cases)
        ]:
            # must contain at least one valid python script
            if any(
                f.endswith(".py") for f in sources_storage.list_folder_files(name, to_root=False)
            ):
                candidates.append(name)

    candidates.sort()
    return candidates


def _get_docstring_for_module(sources_storage: FileStorage, source_name: str) -> str:
    # read the docs
    init_py = os.path.join(source_name, utils.MODULE_INIT)
    docstring: str = ""
    if sources_storage.has_file(init_py):
        docstring = get_module_docstring(sources_storage.load(init_py))
        if docstring:
            docstring = docstring.splitlines()[0]
    return docstring


def get_template_configuration(
    sources_storage: FileStorage, source_name: str, display_source_name: str
) -> SourceConfiguration:
    destination_pipeline_file_name = display_source_name + PIPELINE_FILE_SUFFIX
    source_pipeline_file_name = source_name + PIPELINE_FILE_SUFFIX

    if not sources_storage.has_file(source_pipeline_file_name):
        source_pipeline_file_name = DEFAULT_PIPELINE_TEMPLATE

    docstring = get_module_docstring(sources_storage.load(source_pipeline_file_name))
    if docstring:
        docstring = docstring.splitlines()[0]
    return SourceConfiguration(
        "template",
        source_pipeline_file_name.replace("pipeline.py", ""),
        sources_storage,
        source_pipeline_file_name,
        destination_pipeline_file_name,
        [],
        SourceRequirements([]),
        docstring,
        source_pipeline_file_name == DEFAULT_PIPELINE_TEMPLATE,
    )


def _get_source_files(sources_storage: FileStorage, source_name: str) -> List[str]:
    """Get all files that belong to source `source_name`"""
    files: List[str] = []
    for root, subdirs, _files in os.walk(sources_storage.make_full_path(source_name)):
        # filter unwanted files
        for subdir in list(subdirs):
            if any(fnmatch.fnmatch(subdir, ignore) for ignore in IGNORE_FILES):
                subdirs.remove(subdir)
        rel_root = sources_storage.to_relative_path(root)
        files.extend(
            [
                os.path.join(rel_root, file)
                for file in _files
                if all(not fnmatch.fnmatch(file, ignore) for ignore in IGNORE_FILES)
            ]
        )
    return files


def get_core_source_configuration(
    sources_storage: FileStorage, source_name: str, eject_source: bool
) -> SourceConfiguration:
    src_pipeline_file_root = Path(os.path.dirname(os.path.realpath(__file__))).parent

    src_pipeline_file = src_pipeline_file_root.joinpath(
        TEMPLATES_MODULE_NAME, CORE_SOURCE_TEMPLATE_MODULE_NAME, source_name + PIPELINE_FILE_SUFFIX
    )
    dest_pipeline_file = source_name + PIPELINE_FILE_SUFFIX
    files: List[str] = _get_source_files(sources_storage, source_name) if eject_source else []

    return SourceConfiguration(
        "core",
        "dlt.sources." + source_name,
        sources_storage,
        str(src_pipeline_file),
        dest_pipeline_file,
        files,
        SourceRequirements([]),
        _get_docstring_for_module(sources_storage, source_name),
        False,
    )


def get_verified_source_configuration(
    sources_storage: FileStorage, source_name: str
) -> SourceConfiguration:
    if not sources_storage.has_folder(source_name):
        raise VerifiedSourceRepoError(
            f"Verified source {source_name} could not be found in the repository", source_name
        )
    # find example script
    example_script = f"{source_name}{PIPELINE_FILE_SUFFIX}"
    if not sources_storage.has_file(example_script):
        raise VerifiedSourceRepoError(
            f"Pipeline example script {example_script} could not be found in the repository",
            source_name,
        )
    files = _get_source_files(sources_storage, source_name)
    # read requirements
    requirements_path = os.path.join(source_name, utils.REQUIREMENTS_TXT)
    if sources_storage.has_file(requirements_path):
        requirements = SourceRequirements.from_string(sources_storage.load(requirements_path))
    else:
        requirements = SourceRequirements([])
    # find requirements
    return SourceConfiguration(
        "verified",
        source_name,
        sources_storage,
        example_script,
        example_script,
        files,
        requirements,
        _get_docstring_for_module(sources_storage, source_name),
        False,
    )


def get_core_sources_storage() -> FileStorage:
    """Get FileStorage for core sources"""
    local_path = (
        Path(os.path.dirname(os.path.realpath(__file__))).parent.parent / SOURCES_MODULE_NAME
    )
    return FileStorage(str(local_path))


def get_single_file_templates_storage() -> FileStorage:
    """Get FileStorage for single file templates"""
    # look up init storage in core
    init_path = (
        Path(os.path.dirname(os.path.realpath(__file__))).parent
        / TEMPLATES_MODULE_NAME
        / SINGLE_FILE_TEMPLATE_MODULE_NAME
    )
    return FileStorage(str(init_path))


def gen_index_diff(
    local_index: TVerifiedSourceFileIndex, remote_index: TVerifiedSourceFileIndex
) -> Tuple[
    Dict[str, TVerifiedSourceFileEntry],
    Dict[str, TVerifiedSourceFileEntry],
    Dict[str, TVerifiedSourceFileEntry],
]:
    deleted: Dict[str, TVerifiedSourceFileEntry] = {}
    modified: Dict[str, TVerifiedSourceFileEntry] = {}
    new: Dict[str, TVerifiedSourceFileEntry] = {}

    for name, entry in remote_index["files"].items():
        if name not in local_index["files"]:
            new[name] = entry
        elif entry["sha3_256"] != local_index["files"][name]["sha3_256"]:
            modified[name] = entry

    for name, entry in local_index["files"].items():
        if name not in remote_index["files"]:
            deleted[name] = entry

    return new, modified, deleted


def find_conflict_files(
    local_index: TVerifiedSourceFileIndex,
    remote_new: Dict[str, TVerifiedSourceFileEntry],
    remote_modified: Dict[str, TVerifiedSourceFileEntry],
    remote_deleted: Dict[str, TVerifiedSourceFileEntry],
    dest_storage: FileStorage,
) -> Tuple[List[str], List[str]]:
    """Use files index from .sources to identify modified files via sha3 content hash"""

    conflict_modified: List[str] = []

    def is_file_modified(file: str, entry: TVerifiedSourceFileEntry) -> bool:
        with dest_storage.open_file(file, "rb") as f:
            file_blob = f.read()
        # file exists but was not changed
        return hashlib.sha3_256(file_blob).hexdigest() != entry["sha3_256"]

    for file, entry in remote_new.items():
        if dest_storage.has_file(file):
            # if incoming file is different from local
            if is_file_modified(file, entry):
                conflict_modified.append(file)
        else:
            # file is new local and remote
            pass

    for file, entry in remote_modified.items():
        if dest_storage.has_file(file):
            # if local file was changes and it is different from incoming
            if is_file_modified(file, entry) and is_file_modified(file, local_index["files"][file]):
                conflict_modified.append(file)
        else:
            # file was deleted but is modified on remote
            conflict_modified.append(file)

    conflict_deleted: List[str] = []
    for file, entry in remote_deleted.items():
        if dest_storage.has_file(file):
            if is_file_modified(file, entry):
                conflict_deleted.append(file)
        else:
            # file deleted locally and on remote -> ok
            pass

    return conflict_modified, conflict_deleted



from abc import ABC, abstractmethod
import dataclasses
import contextlib
from threading import BoundedSemaphore
from types import TracebackType
from typing import (
    ClassVar,
    Optional,
    NamedTuple,
    Literal,
    Sequence,
    Iterable,
    Type,
    List,
    ContextManager,
    Dict,
    Any,
    TypeVar,
    Tuple,
)
from typing_extensions import Annotated
import datetime  # noqa: 251

from dlt.common import logger, pendulum
from dlt.common.configuration.specs.base_configuration import extract_inner_hint
from dlt.common.configuration import configspec, NotResolved
from dlt.common.configuration.specs import (
    BaseConfiguration,
    CredentialsConfiguration,
    known_sections,
)
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.destination.utils import (
    resolve_replace_strategy,
    verify_schema_capabilities,
    verify_supported_data_types,
)
from dlt.common.exceptions import TerminalException
from dlt.common.metrics import LoadJobMetrics
from dlt.common.normalizers.naming import NamingConvention

from dlt.common.schema import Schema, TSchemaTables
from dlt.common.schema.typing import (
    C_DLT_LOAD_ID,
    TLoaderReplaceStrategy,
    TTableFormat,
    TTableSchema,
)
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.destination.exceptions import (
    DestinationSchemaTampered,
    DestinationTransientException,
)
from dlt.common.destination.utils import prepare_load_table
from dlt.common.storages import FileStorage
from dlt.common.storages.load_storage import ParsedLoadJobFileName
from dlt.common.storages.load_package import LoadJobInfo, TPipelineStateDoc
from dlt.common.typing import is_optional_type

TDestinationDwhClient = TypeVar("TDestinationDwhClient", bound="DestinationClientDwhConfiguration")

DEFAULT_FILE_LAYOUT = "{table_name}/{load_id}.{file_id}.{ext}"


class StorageSchemaInfo(NamedTuple):
    version_hash: str
    schema_name: str
    version: int
    engine_version: int
    inserted_at: datetime.datetime
    schema: str

    @classmethod
    def from_normalized_mapping(
        cls, normalized_doc: Dict[str, Any], naming_convention: NamingConvention
    ) -> "StorageSchemaInfo":
        """Instantiate this class from mapping where keys are normalized according to given naming convention

        Args:
            normalized_doc: Mapping with normalized keys (e.g. {Version: ..., SchemaName: ...})
            naming_convention: Naming convention that was used to normalize keys

        Returns:
            StorageSchemaInfo: Instance of this class
        """
        return cls(
            version_hash=normalized_doc[naming_convention.normalize_identifier("version_hash")],
            schema_name=normalized_doc[naming_convention.normalize_identifier("schema_name")],
            version=normalized_doc[naming_convention.normalize_identifier("version")],
            engine_version=normalized_doc[naming_convention.normalize_identifier("engine_version")],
            inserted_at=normalized_doc[naming_convention.normalize_identifier("inserted_at")],
            schema=normalized_doc[naming_convention.normalize_identifier("schema")],
        )

    def to_normalized_mapping(self, naming_convention: NamingConvention) -> Dict[str, Any]:
        """Convert this instance to mapping where keys are normalized according to given naming convention

        Args:
            naming_convention: Naming convention that should be used to normalize keys

        Returns:
            Dict[str, Any]: Mapping with normalized keys (e.g. {Version: ..., SchemaName: ...})
        """
        return {
            naming_convention.normalize_identifier(key): value
            for key, value in self._asdict().items()
        }


@dataclasses.dataclass
class StateInfo:
    version: int
    engine_version: int
    pipeline_name: str
    state: str
    created_at: datetime.datetime
    version_hash: Optional[str] = None
    _dlt_load_id: Optional[str] = None

    def as_doc(self) -> TPipelineStateDoc:
        doc: TPipelineStateDoc = dataclasses.asdict(self)  # type: ignore[assignment]
        if self._dlt_load_id is None:
            doc.pop(C_DLT_LOAD_ID)  # type: ignore[misc]
        if self.version_hash is None:
            doc.pop("version_hash")
        return doc

    @classmethod
    def from_normalized_mapping(
        cls, normalized_doc: Dict[str, Any], naming_convention: NamingConvention
    ) -> "StateInfo":
        """Instantiate this class from mapping where keys are normalized according to given naming convention

        Args:
            normalized_doc: Mapping with normalized keys (e.g. {Version: ..., PipelineName: ...})
            naming_convention: Naming convention that was used to normalize keys

        Returns:
            StateInfo: Instance of this class
        """
        return cls(
            version=normalized_doc[naming_convention.normalize_identifier("version")],
            engine_version=normalized_doc[naming_convention.normalize_identifier("engine_version")],
            pipeline_name=normalized_doc[naming_convention.normalize_identifier("pipeline_name")],
            state=normalized_doc[naming_convention.normalize_identifier("state")],
            created_at=normalized_doc[naming_convention.normalize_identifier("created_at")],
            version_hash=normalized_doc.get(naming_convention.normalize_identifier("version_hash")),
            _dlt_load_id=normalized_doc.get(naming_convention.normalize_identifier(C_DLT_LOAD_ID)),
        )


@configspec
class DestinationClientConfiguration(BaseConfiguration):
    destination_type: Annotated[str, NotResolved()] = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )  # which destination to load data to
    credentials: Optional[CredentialsConfiguration] = None
    destination_name: Optional[str] = None  # name of the destination
    environment: Optional[str] = None

    __recommended_sections__: ClassVar[Sequence[str]] = (known_sections.DESTINATION, "")

    def fingerprint(self) -> str:
        """Returns a destination fingerprint which is a hash of selected configuration fields. ie. host in case of connection string"""
        return ""

    def __str__(self) -> str:
        """Return displayable destination location"""
        return str(self.credentials)

    @classmethod
    def credentials_type(
        cls, config: "DestinationClientConfiguration" = None
    ) -> Type[CredentialsConfiguration]:
        """Figure out credentials type, using hint resolvers for dynamic types

        For correct type resolution of filesystem, config should have bucket_url populated
        """
        key = "credentials"
        type_ = cls.get_resolvable_fields()[key]
        if key in cls.__hint_resolvers__ and config is not None:
            try:
                # Type hint for this field is created dynamically
                type_ = cls.__hint_resolvers__[key](config)
            except Exception:
                # we suppress failed hint resolutions
                pass
        return extract_inner_hint(type_)


@configspec
class DestinationClientDwhConfiguration(DestinationClientConfiguration):
    """Configuration of a destination that supports datasets/schemas"""

    dataset_name: Annotated[str, NotResolved()] = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )  # dataset cannot be resolved
    """dataset name in the destination to load data to, for schemas that are not default schema, it is used as dataset prefix"""
    default_schema_name: Annotated[Optional[str], NotResolved()] = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )
    """name of default schema to be used to name effective dataset to load data to"""
    replace_strategy: Optional[TLoaderReplaceStrategy] = None
    """How to handle replace disposition for this destination, uses first strategy from caps if not declared"""
    staging_dataset_name_layout: str = "%s_staging"
    """Layout for staging dataset, where %s is replaced with dataset name. placeholder is optional"""
    enable_dataset_name_normalization: bool = True
    """Whether to normalize the dataset name. Affects staging dataset as well."""
    info_tables_query_threshold: int = 1000
    """Threshold for information schema tables query, if exceeded tables will be filtered in code."""

    def _bind_dataset_name(
        self: TDestinationDwhClient, dataset_name: str, default_schema_name: str = None
    ) -> TDestinationDwhClient:
        """Binds the dataset and default schema name to the configuration

        This method is intended to be used internally.
        """
        self.dataset_name = dataset_name
        self.default_schema_name = default_schema_name
        return self

    def normalize_dataset_name(self, schema: Schema) -> str:
        """Builds full db dataset (schema) name out of configured dataset name and schema name: {dataset_name}_{schema.name}. The resulting name is normalized.

        If default schema name is None or equals schema.name, the schema suffix is skipped.
        """
        dataset_name = self._make_dataset_name(schema.name)
        if not dataset_name:
            return dataset_name
        else:
            return self._normalize_identifier(dataset_name, schema.naming)

    def normalize_staging_dataset_name(self, schema: Schema) -> str:
        """Builds staging dataset name out of dataset_name and staging_dataset_name_layout."""
        if "%s" in self.staging_dataset_name_layout:
            # staging dataset name is never empty, otherwise table names must clash
            dataset_name = self._make_dataset_name(schema.name)
            # fill the placeholder
            dataset_name = self.staging_dataset_name_layout % (dataset_name or "")
        else:
            # no placeholder, then layout is a full name. so you can have a single staging dataset
            dataset_name = self.staging_dataset_name_layout

        return self._normalize_identifier(dataset_name, schema.naming)

    def _normalize_identifier(self, identifier: str, naming: NamingConvention) -> str:
        norm_ident = (
            naming.normalize_table_identifier(identifier)
            if self.enable_dataset_name_normalization
            else identifier
        )
        if norm_ident != identifier:
            logger.warning(
                f"Due to normalization dataset name got changed from '{identifier}' to"
                f" '{norm_ident} which will be used to create db schemas or folders. `dataset_name`"
                " field in the pipeline instance will not be changed. We suggest that you use"
                " dataset names that do not need to be normalized or disable dataset name"
                " normalization via `enable_dataset_name_normalization` on destination"
                " configuration."
            )
        return norm_ident

    @classmethod
    def needs_dataset_name(cls) -> bool:
        """Checks if configuration requires dataset name to be present. Empty datasets are allowed
        ie. for schema-less destinations like weaviate or clickhouse
        """
        fields = cls.get_resolvable_fields()
        dataset_name_type = fields["dataset_name"]
        return not is_optional_type(dataset_name_type)

    def _make_dataset_name(self, schema_name: str) -> str:
        if not schema_name:
            raise ValueError("`schema_name` is `None` or empty")

        # if default schema is None then suffix is not added
        if self.default_schema_name is not None and schema_name != self.default_schema_name:
            return (self.dataset_name or "") + "_" + schema_name
        return self.dataset_name


@configspec
class DestinationClientStagingConfiguration(DestinationClientDwhConfiguration):
    """Configuration of a staging destination, able to store files with desired `layout` at `bucket_url`.

    Also supports datasets and can act as standalone destination.
    """

    as_staging_destination: bool = False
    bucket_url: str = None
    # layout of the destination files
    layout: str = DEFAULT_FILE_LAYOUT


@configspec
class DestinationClientDwhWithStagingConfiguration(DestinationClientDwhConfiguration):
    """Configuration of a destination that can take data from staging destination"""

    staging_config: Optional[DestinationClientStagingConfiguration] = None
    """configuration of the staging, if present, injected at runtime"""
    truncate_tables_on_staging_destination_before_load: bool = True
    """If dlt should truncate the tables on staging destination before loading data."""


TLoadJobState = Literal["ready", "running", "failed", "retry", "completed"]


class LoadJob(ABC):
    """
    A stateful load job, represents one job file
    """

    def __init__(self, file_path: str) -> None:
        self._file_path = file_path
        self._file_name = FileStorage.get_file_name_from_file_path(file_path)
        # NOTE: we only accept a full filepath in the constructor
        assert self._file_name != self._file_path
        self._parsed_file_name = ParsedLoadJobFileName.parse(self._file_name)
        self._started_at: pendulum.DateTime = None
        self._finished_at: pendulum.DateTime = None

    def job_id(self) -> str:
        """The job id that is derived from the file name and does not changes during job lifecycle"""
        return self._parsed_file_name.job_id()

    def file_name(self) -> str:
        """A name of the job file"""
        return self._file_name

    def job_file_info(self) -> ParsedLoadJobFileName:
        return self._parsed_file_name

    @abstractmethod
    def state(self) -> TLoadJobState:
        """Returns current state. Should poll external resource if necessary."""
        pass

    @abstractmethod
    def failed_message(self) -> str:
        """The error message in failed or retry states"""
        pass

    @abstractmethod
    def exception(self) -> BaseException:
        """The exception associated with failed or retry states"""
        pass

    def metrics(self) -> Optional[LoadJobMetrics]:
        """Returns job execution metrics"""
        return LoadJobMetrics(
            self._parsed_file_name.job_id(),
            self._file_path,
            self._parsed_file_name.table_name,
            self._started_at,
            self._finished_at,
            self.state(),
            None,
            self._parsed_file_name.retry_count,
        )


class RunnableLoadJob(LoadJob, ABC):
    """Represents a runnable job that loads a single file

    Each job starts in "running" state and ends in one of terminal states: "retry", "failed" or "completed".
    Each job is uniquely identified by a file name. The file is guaranteed to exist in "running" state. In terminal state, the file may not be present.
    In "running" state, the loader component periodically gets the state via `status()` method. When terminal state is reached, load job is discarded and not called again.
    `exception` method is called to get error information in "failed" and "retry" states.

    The `__init__` method is responsible to put the Job in "running" state. It may raise `LoadClientTerminalException` and `LoadClientTransientException` to
    immediately transition job into "failed" or "retry" state respectively.
    """

    def __init__(self, file_path: str) -> None:
        """
        File name is also a job id (or job id is deterministically derived) so it must be globally unique
        """
        # ensure file name
        super().__init__(file_path)
        self._state: TLoadJobState = "ready"
        self._started_at = pendulum.now()
        self._exception: BaseException = None

        # variables needed by most jobs, set by the loader in set_run_vars
        self._schema: Schema = None
        self._load_table: PreparedTableSchema = None
        self._load_id: str = None
        # set by run_managed method
        self._job_client: "JobClientBase" = None
        self._done_event: BoundedSemaphore = None

    def set_run_vars(self, load_id: str, schema: Schema, load_table: PreparedTableSchema) -> None:
        """
        called by the loader right before the job is run
        """
        self._load_id = load_id
        self._schema = schema
        self._load_table = load_table

    @property
    def load_table_name(self) -> str:
        return self._load_table["name"]

    def run_managed(
        self,
        job_client: "JobClientBase",
        done_event: BoundedSemaphore,
        /,
    ) -> None:
        """
        wrapper around the user implemented run method
        """
        # only jobs that are not running or have not reached a final state
        # may be started
        assert self._state in ("ready", "retry")
        self._job_client = job_client
        self._done_event = done_event

        # filepath is now moved to running
        try:
            self._state = "running"
            self._job_client.prepare_load_job_execution(self)
            self.run()
            self._state = "completed"
        except (TerminalException, AssertionError) as e:
            self._state = "failed"
            self._exception = e
            logger.exception(f"Terminal exception in job {self.job_id()} in file {self._file_path}")
        except (DestinationTransientException, Exception) as e:
            self._state = "retry"
            self._exception = e
            logger.exception(
                f"Transient exception in job {self.job_id()} in file {self._file_path}"
            )
        finally:
            # sanity check
            assert self._state in ("completed", "retry", "failed")
            if self._state != "retry":
                self._finished_at = pendulum.now()
                # wake up waiting threads
                if self._done_event:
                    with contextlib.suppress(ValueError):
                        self._done_event.release()

    @abstractmethod
    def run(self) -> None:
        """
        run the actual job, this will be executed on a thread and should be implemented by the user
        exception will be handled outside of this function
        """
        raise NotImplementedError()

    def state(self) -> TLoadJobState:
        """Returns current state. Should poll external resource if necessary."""
        return self._state

    def failed_message(self) -> str:
        return str(self._exception)

    def exception(self) -> BaseException:
        return self._exception


class FollowupJobRequest:
    """Base class for follow up jobs that should be created"""

    @abstractmethod
    def new_file_path(self) -> str:
        """Path to a newly created temporary job file. If empty, no followup job should be created"""
        pass


class HasFollowupJobs:
    """Adds a trait that allows to create single or table chain followup jobs"""

    def create_followup_jobs(self, final_state: TLoadJobState) -> List[FollowupJobRequest]:
        """Return list of jobs requests for jobs that should be created. `final_state` is state to which this job transits"""
        return []


class JobClientBase(ABC):
    def __init__(
        self,
        schema: Schema,
        config: DestinationClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        self.schema = schema
        self.config = config
        self.capabilities = capabilities

    @abstractmethod
    def initialize_storage(self, truncate_tables: Optional[Iterable[str]] = None) -> None:
        """Prepares storage to be used ie. creates database schema or file system folder. Truncates requested tables."""
        pass

    @abstractmethod
    def is_storage_initialized(self) -> bool:
        """Returns if storage is ready to be read/written."""
        pass

    @abstractmethod
    def drop_storage(self) -> None:
        """Brings storage back into not initialized state. Typically data in storage is destroyed."""
        pass

    def verify_schema(
        self, only_tables: Iterable[str] = None, new_jobs: Iterable[ParsedLoadJobFileName] = None
    ) -> List[PreparedTableSchema]:
        """Verifies schema before loading, returns a list of verified loaded tables."""
        if exceptions := verify_schema_capabilities(
            self.schema,
            self.capabilities,
            self.config.destination_type,
            warnings=False,
        ):
            for exception in exceptions:
                logger.error(str(exception))
            raise exceptions[0]

        prepared_tables = [
            self.prepare_load_table(table_name)
            for table_name in set(
                list(only_tables or []) + self.schema.data_table_names(seen_data_only=True)
            )
        ]
        if exceptions := verify_supported_data_types(
            prepared_tables,
            new_jobs,
            self.capabilities,
            self.config.destination_type,
            warnings=False,
        ):
            for exception in exceptions:
                logger.error(str(exception))
            raise exceptions[0]
        return prepared_tables

    def update_stored_schema(
        self,
        only_tables: Iterable[str] = None,
        expected_update: TSchemaTables = None,
    ) -> Optional[TSchemaTables]:
        """Updates storage to the current schema.

        Implementations should not assume that `expected_update` is the exact difference between destination state and the self.schema. This is only the case if
        destination has single writer and no other processes modify the schema.

        Args:
            only_tables (Sequence[str], optional): Updates only listed tables. Defaults to None.
            expected_update (TSchemaTables, optional): Update that is expected to be applied to the destination
        Returns:
            Optional[TSchemaTables]: Returns an update that was applied at the destination.
        """
        # make sure that schema being saved was not modified from the moment it was loaded from storage
        version_hash = self.schema.version_hash
        if self.schema.is_modified:
            raise DestinationSchemaTampered(
                self.schema.name, version_hash, self.schema.stored_version_hash
            )
        return expected_update

    def prepare_load_table(self, table_name: str) -> PreparedTableSchema:
        """Prepares a table schema to be loaded by filling missing hints and doing other modifications requires by given destination.

        Returns: prepared table, note: table schema for `table_name` is cloned
        """
        return prepare_load_table(
            self.schema.tables, self.schema.get_table(table_name), self.capabilities
        )

    @abstractmethod
    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        """Creates a load job for a particular `table` with content in `file_path`. Table is already prepared to be loaded."""
        pass

    def prepare_load_job_execution(  # noqa: B027, optional override
        self, job: RunnableLoadJob
    ) -> None:
        """Prepare the connected job client for the execution of a load job (used for query tags in sql clients)"""
        pass

    def should_truncate_table_before_load(self, table_name: str) -> bool:
        return self.prepare_load_table(table_name)["write_disposition"] == "replace"

    def create_table_chain_completed_followup_jobs(
        self,
        table_chain: Sequence[PreparedTableSchema],
        completed_table_chain_jobs: Optional[Sequence[LoadJobInfo]] = None,
    ) -> List[FollowupJobRequest]:
        """Creates a list of followup jobs that should be executed after a table chain is completed. Tables are already prepared to be loaded."""
        return []

    @abstractmethod
    def complete_load(self, load_id: str) -> None:
        """Marks the load package with `load_id` as completed in the destination. Before such commit is done, the data with `load_id` is invalid."""
        pass

    @abstractmethod
    def __enter__(self) -> "JobClientBase":
        pass

    @abstractmethod
    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        pass


class WithStateSync(ABC):
    @abstractmethod
    def get_stored_schema(self, schema_name: str = None) -> Optional[StorageSchemaInfo]:
        """Retrieves newest schema with given name from destination storage.

        If no name is provided, the newest schema found is retrieved.
        Returns None if the schema is not found or if the dataset/storage tables
        do not exist (DestinationUndefinedEntity is suppressed by SQL implementations).
        """
        pass

    @abstractmethod
    def get_stored_schema_by_hash(self, version_hash: str) -> StorageSchemaInfo:
        """Retrieves the stored schema by hash.

        Returns None if not found or if the dataset/storage tables do not exist
        (DestinationUndefinedEntity is suppressed by SQL implementations).
        """
        pass

    @abstractmethod
    def get_stored_state(self, pipeline_name: str) -> Optional[StateInfo]:
        """Loads compressed state from destination storage.

        Returns None if no state is found for the given pipeline name.
        Raises DestinationUndefinedEntity if the state or loads tables do not exist
        on the destination (e.g. pipeline never ran).
        """
        pass


class WithStagingDataset(ABC):
    """Adds capability to use staging dataset and request it from the loader"""

    @abstractmethod
    def should_load_data_to_staging_dataset(self, table_name: str) -> bool:
        return False

    @abstractmethod
    def with_staging_dataset(self) -> ContextManager["JobClientBase"]:
        """Executes job client methods on staging dataset"""
        return self  # type: ignore

    @staticmethod
    def create_dataset_names(
        schema: Schema, config: DestinationClientDwhConfiguration
    ) -> Tuple[str, str]:
        """
        Creates regular and staging dataset names for given schema and config.
        Raises a value error if the staging name is same as final dataset name.
        returns (dataset_name, staging_dataset_name)

        """
        dataset_name = config.normalize_dataset_name(schema)
        staging_dataset_name = config.normalize_staging_dataset_name(schema)

        if dataset_name == staging_dataset_name:
            logger.error(
                f"Staging dataset name '{staging_dataset_name}' is the same as final dataset name"
                f" '{dataset_name}'."
            )

            raise ValueError(
                f"The staging dataset name '{staging_dataset_name}' is identical to the final"
                f" dataset name '{dataset_name}'. This configuration will cause data loss because"
                " setup commands will truncate the final dataset when they should only truncate"
                " the staging dataset.\nTo fix this, modify the `staging_dataset_name_layout`"
                " setting in your destination configuration. For more information, see:"
                " https://dlthub.com/docs/dlt-ecosystem/staging#staging-dataset"
            )

        return (dataset_name, staging_dataset_name)


class SupportsStagingDestination(ABC):
    """Adds capability to support a staging destination for the load"""

    def should_load_data_to_staging_dataset_on_staging_destination(self, table_name: str) -> bool:
        """If set to True, and staging destination is configured, the data will be loaded to staging dataset on staging destination
        instead of a regular dataset on staging destination. Currently it is used by Athena Iceberg which uses staging dataset
        on staging destination to copy data to iceberg tables stored on regular dataset on staging destination.
        The default is to load data to regular dataset on staging destination from where warehouses like Snowflake (that have their
        own storage) will copy data.
        """
        return False

    @abstractmethod
    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        """If set to True, data in `table` will be truncated on staging destination (regular dataset). This is the default behavior which
        can be changed with a config flag.
        For Athena + Iceberg this setting is always False - Athena uses regular dataset to store Iceberg tables and we avoid touching it.
        For Athena we truncate those tables only on "replace" write disposition.
        """
        pass

    def should_drop_table_on_staging_destination(self, dropped_table: TTableSchema) -> bool:
        """Tells if `dropped_table` should be dropped on staging destination (regular dataset) in addition to dropping the table on
        final destination. This stays False for all the destinations except Athena, non-iceberg where staging destination
        holds actual data which needs to be deleted.
        Note that `dropped_table` may not longer be present in schema. It is present only if it got recreated.
        """
        return False


class SupportsOpenTables(ABC):
    """Provides access to data stored in one of open table formats (iceberg or delta) and intended to
    be implemented by job clients.

    """

    @abstractmethod
    def get_open_table_catalog(
        self, table_format: TTableFormat, catalog_name: Optional[str] = None
    ) -> Any:
        """Gets the catalog that keeps tables' metadata. Currently only pyiceberg Catalog is supported"""

    @abstractmethod
    def get_open_table_location(self, table_format: TTableFormat, table_name: str) -> str:
        """Computes location in which table is stored which is typically a "folder" with table
        data and metadata. Does not verify if table exists.
        Returns:
            str: fully formed url with table location
        """

    @abstractmethod
    def load_open_table(self, table_format: TTableFormat, table_name: str, **kwargs: Any) -> Any:
        """Loads table `table_name` metadata via catalog or directly and returns populated and authenticated
        table client. Currently pyiceberg Table or DeltaTable is returned.
        * table must be present in schema of job client
        * table must physically exist in storage
        * table may be present in associated catalog and may be automatically registered if destination configuration allows for that
        * otherwise table is not found

        raised DestinationUndefinedEntity if table not found
        """

    @abstractmethod
    def is_open_table(self, table_format: TTableFormat, table_name: str) -> bool:
        """Checks if `table_name` is stored with open table format `table_format`. Does not load table. Does not check if
        table exists
        """



from typing import Optional, Set
from urllib.parse import urlsplit, parse_qsl, urlencode, urlunsplit

SENSITIVE_PARAMS = {
    "api_key",
    "token",
    "key",
    "access_token",
    "apikey",
    "api-key",
    "access-token",
    "secret",
    "password",
    "pwd",
    "client_secret",
}


def sanitize_url(url: str, sensitive: Optional[Set[str]] = None) -> str:
    """Sanitizes a URL by replacing sensitive query parameters with '***'.

    Args:
        url: The URL to sanitize
        sensitive: Set of parameter names to redact (if not provided,
            uses the default set: SENSITIVE_PARAMS)

    Returns:
        str: The sanitized URL
    """
    if not sensitive:
        sensitive = SENSITIVE_PARAMS

    parsed = urlsplit(url)
    qs = parse_qsl(parsed.query, keep_blank_values=True)
    qs = [(k, "***" if k.lower() in {s.lower() for s in sensitive} else v) for k, v in qs]
    new_query = urlencode(qs, safe="*")  # Don't encode asterisks
    return urlunsplit(parsed._replace(query=new_query))



from typing import Callable, Dict, List, Optional, Tuple, Union, Set, Any, Iterable, Literal
import sqlglot
import sqlglot.expressions as sge
from sqlglot.expressions import DataType, DATA_TYPE
from sqlglot.optimizer.scope import build_scope

from dlt.common.utils import without_none
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema.typing import TColumnType, TDataType, TColumnSchema, TTableSchemaColumns
from dlt.common.schema.exceptions import CannotCoerceNullException


TSqlGlotDialect = Literal[
    "athena",
    "bigquery",
    "clickhouse",
    "databricks",
    "doris",
    "drill",
    "druid",
    "duckdb",
    "dune",
    "fabric",
    "hive",
    "materialize",
    "mysql",
    "oracle",
    "postgres",
    "presto",
    "prql",
    "redshift",
    "risingwave",
    "snowflake",
    "spark",
    "spark2",
    "sqlite",
    "starrocks",
    "tableau",
    "teradata",
    "trino",
    "tsql",
]

SQLGLOT_TO_DLT_TYPE_MAP: dict[DataType.Type, TDataType] = {
    # NESTED_TYPES
    DataType.Type.OBJECT: "json",
    DataType.Type.STRUCT: "json",
    DataType.Type.NESTED: "json",
    DataType.Type.ARRAY: "json",
    DataType.Type.JSON: "json",
    # TEXT
    DataType.Type.CHAR: "text",
    DataType.Type.NCHAR: "text",
    DataType.Type.NVARCHAR: "text",
    DataType.Type.TEXT: "text",
    DataType.Type.VARCHAR: "text",
    DataType.Type.NAME: "text",
    # SIGNED_INTEGER
    DataType.Type.BIGINT: "bigint",
    DataType.Type.INT: "bigint",
    DataType.Type.INT128: "bigint",
    DataType.Type.INT256: "bigint",
    DataType.Type.MEDIUMINT: "bigint",
    DataType.Type.SMALLINT: "bigint",
    DataType.Type.TINYINT: "bigint",
    # UNSIGNED_INTEGER
    DataType.Type.UBIGINT: "bigint",
    DataType.Type.UINT: "bigint",
    DataType.Type.UINT128: "bigint",
    DataType.Type.UINT256: "bigint",
    DataType.Type.UMEDIUMINT: "bigint",
    DataType.Type.USMALLINT: "bigint",
    DataType.Type.UTINYINT: "bigint",
    # other INTEGER
    DataType.Type.BIT: "bigint",
    # FLOAT
    DataType.Type.DOUBLE: "double",
    DataType.Type.FLOAT: "double",
    # DECIMAL
    DataType.Type.BIGDECIMAL: "decimal",
    DataType.Type.DECIMAL: "decimal",
    DataType.Type.MONEY: "decimal",
    DataType.Type.SMALLMONEY: "decimal",
    DataType.Type.UDECIMAL: "decimal",
    # TEMPORAL
    DataType.Type.DATE: "date",
    DataType.Type.DATE32: "date",
    DataType.Type.DATETIME: "date",
    DataType.Type.DATETIME64: "date",
    DataType.Type.TIMESTAMP: "timestamp",
    DataType.Type.TIMESTAMPLTZ: "timestamp",
    DataType.Type.TIMESTAMPTZ: "timestamp",
    DataType.Type.TIMESTAMP_MS: "timestamp",
    DataType.Type.TIMESTAMP_NS: "timestamp",
    DataType.Type.TIMESTAMP_S: "timestamp",
    DataType.Type.TIME: "time",
    DataType.Type.TIMETZ: "time",
    # binary
    DataType.Type.VARBINARY: "binary",
    # BOOLEAN
    DataType.Type.BOOLEAN: "bool",
    # UNKNOWN
    DataType.Type.UNKNOWN: None,
}

# these types were introduced after our sqlglot minimum version of 23.6.3
try:
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.UDOUBLE] = "decimal"
except AttributeError:
    pass

try:
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.DATETIME2] = "date"
except AttributeError:
    pass

try:
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.SMALLDATETIME] = "date"
except AttributeError:
    pass

try:
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.UNION] = "json"
except AttributeError:
    pass

try:
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.LIST] = "json"
except AttributeError:
    pass

try:
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.VECTOR] = "json"
except AttributeError:
    pass

try:
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.DECIMAL32] = "decimal"
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.DECIMAL64] = "decimal"
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.DECIMAL128] = "decimal"
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.DECIMAL256] = "decimal"
except AttributeError:
    pass


# NOTE in Snowflake, TIMESTAMPNTZ == DATETIME; is this true for dlt?
SQLGLOT_HAS_TIMEZONE = {
    DataType.Type.TIMESTAMP: None,  # default value; False
    DataType.Type.TIMESTAMPLTZ: True,
    DataType.Type.TIMESTAMPTZ: True,
    DataType.Type.TIMESTAMP_MS: False,
    DataType.Type.TIMESTAMP_NS: False,
    DataType.Type.TIMESTAMP_S: False,
    DataType.Type.TIME: False,  # default value; False
    DataType.Type.TIMETZ: True,
}

has_timestampntz = hasattr(DataType.Type, "TIMESTAMPNTZ")
if has_timestampntz:
    SQLGLOT_HAS_TIMEZONE[DataType.Type.TIMESTAMPNTZ] = False
    SQLGLOT_TO_DLT_TYPE_MAP[DataType.Type.TIMESTAMPNTZ] = "timestamp"

DLT_TO_SQLGLOT = {
    "json": DataType.Type.JSON,
    "text": DataType.Type.TEXT,
    "double": DataType.Type.DOUBLE,
    "bool": DataType.Type.BOOLEAN,
    "date": DataType.Type.DATE,
    "timestamp": DataType.Type.TIMESTAMP,
    "bigint": DataType.Type.BIGINT,
    "binary": DataType.Type.VARBINARY,
    "time": DataType.Type.TIME,
    "decimal": DataType.Type.DECIMAL,
}


# TODO should we raise errors on type conversion or silently convert to correct values?
def from_sqlglot_type(sqlglot_type: DATA_TYPE) -> TColumnType:
    """Convert a SQLGlot DataType to dlt column hints.

    reference: https://dlthub.com/docs/general-usage/schema/#tables-and-columns
    """
    if isinstance(sqlglot_type, str):
        sqlglot_type = sge.DataType.build(sqlglot_type)
    elif isinstance(sqlglot_type, sge.DataType.Type):
        sqlglot_type = sge.DataType.build(sqlglot_type)
    assert isinstance(sqlglot_type, sge.DataType)

    # only the `DataType.UNKNOWN` will produce `dlt_type=None`
    dlt_type = SQLGLOT_TO_DLT_TYPE_MAP.get(sqlglot_type.this, "text")

    if dlt_type == "bigint":
        hints = _from_integer_type(sqlglot_type)
    elif dlt_type == "decimal":
        hints = _from_decimal_type(sqlglot_type)
    elif dlt_type == "timestamp":
        hints = _from_timezone_type(sqlglot_type)
    elif dlt_type == "time":
        hints = _from_time_type(sqlglot_type)
    elif dlt_type in ("text", "binary"):
        hints = _from_string_type(sqlglot_type)
    else:
        hints = {}

    # `nullable` is not a required arg on exp.DataType
    nullable = sqlglot_type.args.get("nullable")

    return without_none({"data_type": dlt_type, "nullable": nullable, **hints})  # type: ignore[return-value]


def _from_integer_type(sqlglot_type: sge.DataType) -> TColumnSchema:
    if sqlglot_type.expressions:  # from parameterized type
        assert len(sqlglot_type.expressions) == 1
        assert isinstance(sqlglot_type.expressions[0], sge.DataTypeParam)
        return {"precision": sqlglot_type.expressions[0].this.to_py()}

    return {}


def _from_decimal_type(sqlglot_type: sge.DataType) -> TColumnSchema:
    hints = {}
    if sqlglot_type.expressions:
        assert all(isinstance(e, sge.DataTypeParam) for e in sqlglot_type.expressions)
        assert len(sqlglot_type.expressions) in (1, 2)
        if len(sqlglot_type.expressions) == 1:
            precision = sqlglot_type.expressions[0].this.to_py()
            hints = {"precision": precision}
        elif len(sqlglot_type.expressions) == 2:
            precision = sqlglot_type.expressions[0].this.to_py()
            scale = sqlglot_type.expressions[1].this.to_py()
            hints = {"precision": precision, "scale": scale}
        else:
            raise RuntimeError(
                "Expected 1 or 2 `DataTypeParam` attached to expression. "
                f"Found {len(sqlglot_type.expressions)}: {sqlglot_type.expressions}"
            )

    return hints  # type: ignore[return-value]


def _from_timezone_type(sqlglot_type: sge.DataType) -> TColumnSchema:
    timezone = SQLGLOT_HAS_TIMEZONE.get(sqlglot_type.this)
    precision = None

    if sqlglot_type.expressions:  # from parameterized type
        assert len(sqlglot_type.expressions) == 1
        assert isinstance(sqlglot_type.expressions[0], sge.DataTypeParam)
        precision = sqlglot_type.expressions[0].this.to_py()

    return {"precision": precision, "timezone": timezone}


def _from_time_type(sqlglot_type: sge.DataType) -> TColumnSchema:
    timezone = SQLGLOT_HAS_TIMEZONE.get(sqlglot_type.this)

    if sqlglot_type.expressions:  # from parameterized type
        assert len(sqlglot_type.expressions) == 1
        assert isinstance(sqlglot_type.expressions[0], sge.DataTypeParam)
        precision = sqlglot_type.expressions[0].this.to_py()
        hints = {"precision": precision, "timezone": timezone}
    else:  # from named type
        hints = {"timezone": timezone}

    return hints  # type: ignore[return-value]


def _from_string_type(sqlglot_type: sge.DataType) -> TColumnSchema:
    if sqlglot_type.expressions:  # from parameterized type
        assert len(sqlglot_type.expressions) == 1
        assert isinstance(sqlglot_type.expressions[0], sge.DataTypeParam)
        precision = sqlglot_type.expressions[0].this.to_py()
        hints: TColumnSchema = {"precision": precision}
    else:
        hints = {}
    return hints


def to_sqlglot_type(
    dlt_type: TDataType,
    precision: Optional[int] = None,
    scale: Optional[int] = None,
    timezone: Optional[bool] = None,
    nullable: Optional[bool] = None,
    use_named_types: bool = False,
) -> DataType:
    """Convert the dlt `data_type` and column hints to a SQLGlot DataType expression.

    `if use_named_type is True:` use "named" types with fallback on "parameterized" types.
    `else:` use "parameterized" types everywhere.

    Named types:
    - have some set attributes, e.g., `DataType.Type.DECIMAL64` has `precision=16` and `scale=4`.
    - can be referenced directly in SQL (table definition, CAST(), which could make SQLGlot transpiling more reliable.
    - may not exist in all dialects and will be casted automatically during transpiling
    - can have limited expressivity, e.g., named types for timestamps only included precision 0, 3, 9

    Parameterized types:
    - instead of DECIMAL64, a generic `DataType.Type.DECIMAL` is used with `DataTypeParam` expressions attached
      to represent `precision` and `scale`. SQLGlot is responsible for properly compiling and transpiling them.
    - parameters might be handled differently across dialects and would require greater testing on the dlt side.

    reference: https://dlthub.com/docs/general-usage/schema/#tables-and-columns
    """
    if use_named_types is True:
        try:
            sqlglot_type = _build_named_sqlglot_type(
                dlt_type=dlt_type,
                nullable=nullable,
                precision=precision,
                scale=scale,
                timezone=timezone,
            )
        except TerminalValueError:
            sqlglot_type = _build_parameterized_sqlglot_type(
                dlt_type=dlt_type,
                nullable=nullable,
                precision=precision,
                scale=scale,
                timezone=timezone,
            )
    else:
        sqlglot_type = _build_parameterized_sqlglot_type(
            dlt_type=dlt_type,
            nullable=nullable,
            precision=precision,
            scale=scale,
            timezone=timezone,
        )

    return sqlglot_type


# NOTE Let's ignore named types for now
def _build_named_sqlglot_type(
    dlt_type: TDataType,
    precision: Optional[int] = None,
    scale: Optional[int] = None,
    timezone: Optional[bool] = None,
    nullable: Optional[bool] = None,
) -> DataType:
    hints = {}
    if nullable is not None:
        hints["nullable"] = nullable

    if dlt_type == "bigint" and precision is not None:
        sqlglot_type = _to_named_integer_type(precision=precision)
    elif dlt_type == "decimal" and precision is not None:
        sqlglot_type = _to_named_decimal_type(precision=precision, scale=scale)
    elif dlt_type == "timestamp" and (precision is not None or timezone is not None):
        sqlglot_type = _to_named_timestamp_type(precision=precision, timezone=timezone)
    elif dlt_type == "time" and timezone is not None:
        sqlglot_type = _to_named_time_type(timezone=timezone)
    else:
        raise TerminalValueError()

    return DataType.build(dtype=sqlglot_type, **hints)  # type: ignore[arg-type]


def _build_parameterized_sqlglot_type(
    dlt_type: TDataType,
    nullable: Optional[bool] = None,
    precision: Optional[int] = None,
    scale: Optional[int] = None,
    timezone: Optional[bool] = None,
) -> DataType:
    hints: TColumnType = {}
    # `nullable` should always be added to DataType
    if nullable is not None:
        hints["nullable"] = nullable

    if dlt_type == "bigint" and precision is not None:
        # NOTE INT precision is not supported by many backend.
        # For instance, INT(5) in MySQL should be DECIMAL(5,0) in DuckDB
        # Currently, SQLGLot can't transpile from INT(5) to DECIMAL(5,0).
        # Using non-parameterized type would be more reliable in these cases.
        sqlglot_type = sge.DataType.build(f"INT({precision})", **hints)

    elif dlt_type == "decimal" and precision is not None:
        # this code is inspired from sqlglot.dialects.dialect.build_default_decimal_type
        params = f"({precision}{f', {scale}' if scale is not None else ''})"
        sqlglot_type = sge.DataType.build(f"DECIMAL{params}", **hints)

    elif dlt_type == "timestamp" and (precision is not None or timezone is not None):
        base_sqlglot_type = _to_named_timestamp_type(precision=None, timezone=timezone)
        params = f"({precision})" if precision is not None else ""
        sqlglot_type = sge.DataType.build(f"{base_sqlglot_type.value}{params}", **hints)

    elif dlt_type == "time" and (precision is not None or timezone is not None):
        base_sqlglot_type = _to_named_time_type(timezone=timezone)
        params = f"({precision})" if precision is not None else ""
        sqlglot_type = sge.DataType.build(f"{base_sqlglot_type.value}{params}", **hints)

    elif dlt_type == "text" and precision is not None:
        sqlglot_type = sge.DataType.build(f"TEXT({precision})", **hints)

    elif dlt_type == "binary" and precision is not None:
        sqlglot_type = sge.DataType.build(f"VARBINARY({precision})", **hints)
    elif dlt_type == "wei":
        sqlglot_type = sge.DataType.build("DECIMAL(38, 0)", **hints)
    else:
        sqlglot_type = sge.DataType.build(DLT_TO_SQLGLOT[dlt_type], **hints)

    return sqlglot_type


def _to_named_integer_type(precision: Optional[int]) -> DataType.Type:
    """Return the SQLGlot INTEGER DataType with precision `>= precision`."""
    if precision is None:
        return DataType.Type.BIGINT
    elif precision <= 0:
        # raise TerminalValueError(f"Precision must be positive, got {precision}")
        sqlglot_type = DataType.Type.BIGINT  # default value
    elif precision <= 3:
        sqlglot_type = DataType.Type.TINYINT
    elif precision <= 5:
        sqlglot_type = DataType.Type.SMALLINT
    elif precision <= 8:
        sqlglot_type = DataType.Type.MEDIUMINT
    elif precision <= 10:
        sqlglot_type = DataType.Type.INT
    elif precision <= 19:
        sqlglot_type = DataType.Type.BIGINT
    elif precision <= 39:
        sqlglot_type = DataType.Type.INT128
    elif precision <= 78:
        sqlglot_type = DataType.Type.INT256
    else:
        raise TerminalValueError(
            f"Precision must be less than maximum precision of 78, got {precision}"
        )

    return sqlglot_type


# TODO
def _to_named_decimal_type(precision: Optional[int], scale: Optional[int]) -> DataType.Type:
    sqlglot_type = DataType.Type.DECIMAL
    # if precision is None:
    #     sqlglot_type = DataType.Type.DECIMAL
    # elif precision <= 0:
    #     raise TerminalValueError(f"Precision must be positive, got {precision}")
    # else

    return sqlglot_type


def _to_named_timestamp_type(precision: Optional[int], timezone: Optional[bool]) -> DataType.Type:
    if precision is None:
        if timezone is True:
            sqlglot_type = DataType.Type.TIMESTAMPTZ
        elif timezone is False and has_timestampntz:
            sqlglot_type = DataType.Type.TIMESTAMPNTZ
        elif timezone is None:
            sqlglot_type = DataType.Type.TIMESTAMP
        else:
            raise TerminalValueError(
                f"Received `timezone` value not in (True, False, None), got {timezone}"
            )

    else:
        if timezone is None and precision == 0:
            sqlglot_type = sge.DataType.Type.TIMESTAMP_S
        elif timezone is None and precision <= 3:
            sqlglot_type = sge.DataType.Type.TIMESTAMP_MS
        elif timezone is None and precision <= 9:
            sqlglot_type = sge.DataType.Type.TIMESTAMP_NS
        else:
            raise TerminalValueError(
                f"Precision must be less than maximum precision of 9, got {precision}"
            )

    return sqlglot_type


def _to_named_time_type(timezone: Optional[bool]) -> DataType.Type:
    if timezone is None:
        sqlglot_type = DataType.Type.TIME
    elif timezone is True:
        sqlglot_type = DataType.Type.TIMETZ
    elif timezone is False:
        sqlglot_type = DataType.Type.TIME
    else:
        raise TerminalValueError(
            f"Received `timezone` value not in (True, False, None), got {timezone}"
        )

    return sqlglot_type


# TODO implement an SQLGlot annotator for setting dlt hints instead of relying on types.
# ref: https://github.com/tobymao/sqlglot/blob/17f7eaff564790b1fe7faa414143accf362f550e/sqlglot/optimizer/annotate_types.py#L135
def set_metadata(sqlglot_type: sge.DataType, metadata: TColumnSchema) -> sge.DataType:
    """Set a metadata dictionary on the SQGLot DataType object.

    By attaching dlt hints to a DataType object, they will be propagated
    until the DataType is modified.
    """
    sqlglot_type._meta = _filter_dlt_hints(metadata)  # type: ignore[assignment]
    return sqlglot_type


def get_metadata(sqlglot_type: sge.DataType) -> TColumnSchema:
    """Get a metadata dictionary from the SQLGlot DataType object."""
    metadata = sqlglot_type.meta
    return _filter_dlt_hints(metadata)  # type: ignore[arg-type]


def _filter_dlt_hints(hints: TColumnSchema) -> TColumnSchema:
    """Filter the metadata dictionary to only include dlt hints."""
    DATA_TYPE_HINTS = ("data_type", "nullable", "precision", "scale", "timezone")
    # allow original name to be propagated, so we can track it if aliases were used
    ALLOWED_HINTS = ("name", "hard_delete", "incremental")
    # only propagate specially designated hints
    # TODO review all x- hints we use and decide which to propagate
    CUSTOM_HINT_PREFIX = "x-annotation"

    final_hints = {}
    for k, v in hints.items():
        # those are directly expressed via the DataType and Column
        if k in DATA_TYPE_HINTS:
            continue

        if k in ALLOWED_HINTS:
            final_hints[k] = v
        elif k.startswith(CUSTOM_HINT_PREFIX):
            final_hints[k] = v

    return final_hints  # type: ignore[return-value]


def query_is_complex(
    parsed_select: Union[sge.Select, sge.Union],
    columns: Set[str],
) -> bool:
    """
    Return **True** unless the query is provably “simple”.
    A *simple* query
    1. references **exactly one** physical table,
    2. contains no complex constructs (CTEs, sub-queries, derived tables,
       unions, window functions, GROUP BY, DISTINCT, etc.),
    3. projects either
         • a plain/qualified star with only constant literals after it, or
         • the full, explicit list of *all* columns with only constant
           literals after it.
    Anything we cannot *prove* to be simple is conservatively flagged
    as complex.
    """
    root_scope = build_scope(parsed_select)
    tables: list[sge.Table] = []
    non_literal_cols: dict[str, int] = {}
    star_present = False
    for node in root_scope.walk():
        if isinstance(node, sge.Table):
            tables.append(node)
        if isinstance(node, sge.Window):
            return True
        if isinstance(node, sge.SetOperation):
            return True
        if isinstance(node, sge.With):
            return True
        if isinstance(node, sge.Distinct):
            return True
        if isinstance(node, sge.Group):
            return True
        if isinstance(node, sge.DPipe):
            return True
        # Detect a star in the outermost projection list.
        # A star can be either “*” or a qualified “t.*”.
        # In the AST:
        #   • bare  “*”    Star -> SELECT
        #   • qual. “t.*”  Star -> Column -> SELECT
        if isinstance(node, sge.Star) and (
            node.parent == parsed_select or (node.parent and node.parent.parent == parsed_select)
        ):
            star_present = True
        # Detect plain (aliased) column references that belong to the outermost projection list.
        if isinstance(node, sge.Column) and not isinstance(node.this, sge.Star):
            # it's an (aliased) column in our outer select
            if node.parent == parsed_select or (
                isinstance(node.parent, sge.Alias) and node.parent_select == parsed_select
            ):
                non_literal_cols[node.name] = non_literal_cols.get(node.name, 0) + 1
        # An aliased expression that is *not* just a column or a
        # literal disqualifies the query from being “simple
        if isinstance(node, sge.Alias):
            if not (isinstance(node.this, sge.Column) or isinstance(node.this, sge.Literal)):
                return True
    if len(tables) > 1:
        return True
    if star_present and not non_literal_cols:
        return False
    non_literal_cols_names = set(non_literal_cols.keys())
    # if any column is selected twice, it is complex
    if any(count > 1 for count in non_literal_cols.values()):
        return True
    # if selected columns are exactly the same as the table columns, it is simple
    if non_literal_cols_names == columns:
        return False
    # if newly added columns are all exclusively dlt columns, it is simple
    if (
        all(col_name.startswith("_dlt_") for col_name in non_literal_cols_names - columns)
        and non_literal_cols_names - columns != set()
    ):
        return False
    # if missing columns are all exclusively dlt columns, it is simple
    if (
        all(col_name.startswith("_dlt_") for col_name in columns - non_literal_cols_names)
        and columns - non_literal_cols_names != set()
    ):
        return False
    return True


def build_typed_literal(
    value: Any, sqlglot_type: sge.DataType = None
) -> Union[sge.Expression, sge.Tuple]:
    """Create a literal and CAST it to the requested sqlglot DataType."""

    def _literal(v: Any) -> sge.Expression:
        lit: sge.Expression
        if v is None:
            lit = sge.Null()
        elif isinstance(v, str):
            lit = sge.Literal.string(v)
        elif isinstance(v, (int, float)):
            lit = sge.Literal.number(v)
        elif isinstance(v, (bytes, bytearray)):
            lit = sge.Literal.string(v.hex())
        else:
            lit = sge.Literal.string(str(v))
        return sge.Cast(this=lit, to=sqlglot_type.copy()) if sqlglot_type is not None else lit

    if isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
        return sge.Tuple(expressions=[_literal(v) for v in value])
    else:
        return _literal(value)


DLT_SUBQUERY_NAME = "_dlt_subquery"


class SqlModel:
    """A SqlModel is a named tuple that contains a query and a dialect.
    It is used to represent a SQL query and the dialect to use for parsing it.
    """

    __slots__ = ("_query", "_dialect")

    def __init__(self, query: str, dialect: Optional[str] = None) -> None:
        self._query = query
        self._dialect = dialect

    def to_sql(self) -> str:
        return self._query

    @property
    def query_dialect(self) -> str:
        return self._dialect

    @classmethod
    def from_query_string(cls, query: str, dialect: Optional[str] = None) -> "SqlModel":
        """Creates a SqlModel from a raw SQL query string using sqlglot.
        Ensures that the parsed query is an instance of sqlglot.exp.Select.

        Args:
            query: The raw SQL query string.
            dialect: The SQL dialect to use for parsing.

        Returns:
            An instance of SqlModel with the normalized query and dialect.

        Raises:
            ValueError: If the parsed query is not an instance of sqlglot.exp.Select.
        """
        parsed_query = sqlglot.parse_one(query, read=dialect)

        if not isinstance(parsed_query, sge.Select):
            raise ValueError("Only SELECT statements are allowed to create a `SqlModel`.")

        normalized_query = parsed_query.sql(dialect=dialect)
        return cls(query=normalized_query, dialect=dialect)


def uuid_expr_for_dialect(dialect: TSqlGlotDialect, load_id: str) -> sge.Expression:
    """Generates a UUID expression based on the specified dialect.

    Args:
        dialect: The SQL dialect for which the UUID expression needs to be generated.
        load_id: The load ID used for deterministic UUID generation (redshift).

    Returns:
        A SQL expression that generates a UUID for the specified dialect.
    """
    # NOTE: redshift and sqlite don't have an in-built uuid function
    if dialect == "redshift":
        row_num = sge.Window(
            this=sge.Anonymous(this="row_number"),
            partition_by=None,
            order=None,
        )
        concat_expr = sge.func(
            "CONCAT",
            sge.Literal.string(load_id),
            sge.Literal.string("-"),
            row_num,
        )
        return sge.func("MD5", concat_expr)
    elif dialect == "sqlite":
        return sge.func(
            "lower",
            sge.func("hex", sge.func("randomblob", sge.Literal.number(16))),
        )
    elif dialect == "clickhouse":
        return sge.func("generateUUIDv4")
    # NOTE: UUID in Athena creates a native UUID data type
    # which needs to be typecasted
    elif dialect == "athena":
        return sge.Cast(this=sge.func("UUID"), to=sge.DataType.build("VARCHAR"))
    else:
        return sge.func("UUID")


def get_select_column_names(
    selects: List[sge.Expression],
    dialect: Optional[TSqlGlotDialect] = None,
) -> List[str]:
    """Extract output column names from a SELECT clause's expression list.

    Handles Alias (returns alias), Column and Dot (returns output_name with
    fallback to name — needed for BigQuery quoted identifiers that parse as Dot).
    Raises ValueError for star expressions or unsupported expression types.

    Args:
        selects: The ``.selects`` list from a parsed SELECT statement.
        dialect: SQL dialect used only for error message formatting.

    Returns:
        Column output names in SELECT order.
    """
    names: List[str] = []
    for select in selects:
        if isinstance(select, sge.Star):
            raise ValueError(
                "Star expressions are not supported."
                " Please rewrite the query to explicitly specify columns."
            )
        if isinstance(select, sge.Alias):
            names.append(select.alias)
        elif isinstance(select, sge.Column) or isinstance(select, sge.Dot):
            if isinstance(select.this, sge.Star):
                raise ValueError(
                    "Star expressions are not supported."
                    " Please rewrite the query to explicitly specify columns."
                )
            name = select.output_name or select.name
            names.append(name)
        else:
            sql_text = select.sql(dialect) if dialect else str(select)
            raise ValueError(
                "\n\nUnsupported SELECT expression in the model query:\n\n"
                f"  {sql_text}\n\nOnly simple column selections like `column`"
                " or `column AS alias` are currently supported.\n"
            )
    return names


def filter_select_column_names(
    selects: List[sge.Expression],
    discard_columns: Set[str],
    normalize_fn: Callable[[str], str],
    dialect: Optional[TSqlGlotDialect] = None,
) -> List[str]:
    """Return SELECT column names excluding those whose normalized form is in discard_columns.

    Preserves original SELECT order.

    Args:
        selects: The ``.selects`` list from a parsed SELECT statement.
        discard_columns: Set of normalized column names to exclude.
        normalize_fn: Callable that normalizes a column name (e.g. casefold).
        dialect: SQL dialect used only for error message formatting.

    Returns:
        Remaining column names in their original SELECT order.
    """
    all_names = get_select_column_names(selects, dialect)
    return [name for name in all_names if normalize_fn(name) not in discard_columns]


def validate_no_star_select(parsed_select: sge.Select, dialect: TSqlGlotDialect) -> None:
    """Raises ValueError if the SELECT statement contains a star expression.

    Args:
        parsed_select: The parsed SELECT statement.
        dialect: The SQL dialect (used for error message formatting).
    """
    if any(
        isinstance(expr, sge.Star)
        or (isinstance(expr, sge.Column) and isinstance(expr.this, sge.Star))
        for expr in parsed_select.selects
    ):
        raise ValueError(
            "\n\nA `SELECT *` was detected in the model query:\n\n"
            f"{parsed_select.sql(dialect)}\n\n"
            "Model queries using a star (`*`) expression cannot be normalized. "
            "Please rewrite the query to explicitly specify the columns to be selected.\n"
        )


def build_outer_select_statement(
    select_dialect: TSqlGlotDialect,
    parsed_select: sge.Select,
    columns: TTableSchemaColumns,
    normalize_casefold_fn: Callable[[str], str],
) -> Tuple[sge.Select, bool]:
    """Wraps the parsed SELECT in a subquery and builds an outer SELECT statement.

    Args:
        select_dialect: The SQL dialect to use for parsing and formatting.
        parsed_select: The parsed SELECT statement.
        columns: The schema columns to match.
        normalize_casefold_fn: A callable that normalizes and casefolds an identifier.

    Returns:
        Tuple of the outer SELECT statement and a flag indicating if reordering is needed.
    """
    subquery = parsed_select.subquery(alias=DLT_SUBQUERY_NAME)

    column_names = get_select_column_names(parsed_select.selects, select_dialect)

    selected_columns: List[str] = []
    outer_selects: List[sge.Expression] = []
    for name in column_names:
        norm_casefolded = normalize_casefold_fn(name)
        selected_columns.append(norm_casefolded)

        column_ref = sge.Dot(
            this=sqlglot.to_identifier(DLT_SUBQUERY_NAME),
            expression=sqlglot.to_identifier(name, quoted=True),
        )

        outer_selects.append(column_ref.as_(norm_casefolded, quoted=True))

    needs_reordering = selected_columns != list(columns.keys())

    outer_select = sqlglot.select(*outer_selects).from_(subquery)

    return outer_select, needs_reordering


def reorder_or_adjust_outer_select(
    outer_parsed_select: sge.Select,
    columns: TTableSchemaColumns,
    normalize_casefold_fn: Callable[[str], str],
    schema_name: str,
    table_name: str,
) -> None:
    """Reorders or adjusts the SELECT statement to match the schema.

    Adds missing columns as NULL and removes extra columns not in the schema.

    Args:
        outer_parsed_select: The parsed outer SELECT statement.
        columns: The schema columns to match.
        normalize_casefold_fn: A callable that normalizes and casefolds an identifier.
        schema_name: The schema name (used for error messages).
        table_name: The table name (used for error messages).
    """
    alias_map = {expr.alias.lower(): expr for expr in outer_parsed_select.selects}

    new_selects: List[sge.Expression] = []
    for col in columns:
        lower_col = col.lower()
        expr = alias_map.get(lower_col)
        if expr:
            new_selects.append(expr)
        else:
            if columns[col]["nullable"]:
                new_selects.append(
                    sge.Alias(
                        this=sge.Null(),
                        alias=sqlglot.to_identifier(normalize_casefold_fn(col), quoted=True),
                    )
                )
            else:
                exc = CannotCoerceNullException(schema_name, table_name, col)
                exc.args = (
                    exc.args[0]
                    + " — column is missing from the SELECT clause but is non-nullable and must"
                    " be explicitly selected",
                )
                raise exc

    outer_parsed_select.set("expressions", new_selects)


def normalize_query_identifiers(
    query: sge.Query,
    naming_convention: Any,
) -> sge.Query:
    """Normalize all logical identifiers in a query to match the naming convention.

    Normalizes table names, column names, and aliases. Does not modify database
    or catalog names. Call before bind_query to ensure identifiers match the
    dlt schema naming.

    Args:
        query: Query with logical (possibly unnormalized) identifiers
        naming_convention: Naming convention to apply

    Returns:
        Query with all identifiers normalized
    """
    query = query.copy()

    for node in query.walk():
        # normalize table names
        if isinstance(node, sge.Table):
            normalized = naming_convention.normalize_tables_path(node.name)
            if node.name != normalized:
                node.this.set("this", normalized)

        # normalize column and other identifiers (but not table/db/catalog)
        elif isinstance(node, sge.Identifier):
            # skip identifiers that are part of Table nodes (table/db/catalog names)
            if isinstance(node.parent, sge.Table):
                continue

            # table qualifiers in column references must use normalize_tables_path
            # to stay consistent with table names in FROM/JOIN
            if isinstance(node.parent, sge.Column) and node.arg_key == "table":
                normalized = naming_convention.normalize_tables_path(node.this)
            else:
                normalized = naming_convention.normalize_path(node.this)
            if node.this != normalized:
                node.set("this", normalized)

        # normalize aliases
        elif isinstance(node, sge.Alias) and node.alias:
            normalized = naming_convention.normalize_path(node.alias)
            if node.alias != normalized:
                node.set("alias", sqlglot.to_identifier(normalized, quoted=False))

    return query


def bind_query(
    qualified_query: sge.Query,
    sqlglot_schema: Any,  # SQLGlotSchema
    *,
    expand_table_name: Callable[[str], List[str]],
    casefold_identifier: Callable[[str], str],
) -> sge.Query:
    """Binds a logical query (compliant with dlt schema) to physical tables in the destination dataset.

    This function performs name resolution/binding - a standard query processing step that maps
    logical identifiers to actual database objects. It takes a query using dlt's logical naming
    convention and adapts it for execution on the destination database.
    Note that outer select list after binding preserve select list in `qualified_query`.

    Binding steps performed:
    1. **Table name expansion**: Resolves table references to fully qualified physical paths
       (e.g., for ClickHouse: `dataset.table` -> `dataset___table`)
    2. **Identifier case-folding**: Applies destination-specific case transformation
       (`str.upper` for Snowflake, `str.lower` for most databases, `str` for case-sensitive)
    3. **Identifier quoting**: Quotes all identifiers for safe execution and to preserve
       exact casing (e.g., `column_name` -> `"column_name"`)
    4. **Alias preservation**: For case-folding destinations, adds aliases to SELECT columns
       to maintain compatibility with dlt schema naming (e.g., `SELECT "VALUE" AS "value"`)

    Args:
        qualified_query: SQLGlot query expression with qualified table/column references
        sqlglot_schema: Schema mapping for name validation and column resolution
        expand_table_name: Function that expands table name to fully qualified path [catalog, schema, table]
        casefold_identifier: Case transformation function (`str`, `str.upper`, or `str.lower`)

    Returns:
        Bound query expression ready for execution on the destination database
    """
    qualified_query = qualified_query.copy()
    is_casefolding = casefold_identifier is not str

    # preserve "column" names in original selects which are done in dlt schema namespace
    orig_selects: Dict[int, str] = None
    if is_casefolding:
        orig_selects = {}
        for i, proj in enumerate(qualified_query.selects):
            orig_selects[i] = proj.name or proj.args["alias"].name

    # case fold all identifiers and quote
    for node in qualified_query.walk():
        if isinstance(node, sge.Table):
            # expand named of known tables. this is currently clickhouse things where
            # we use dataset.table in queries but render those as dataset___table
            if sqlglot_schema.column_names(node):
                expanded_path = expand_table_name(node.name)
                # set the table name
                if node.name != expanded_path[-1]:
                    node.this.set("this", expanded_path[-1])
                # set the dataset/schema name
                if node.db != expanded_path[-2]:
                    node.set("db", sqlglot.to_identifier(expanded_path[-2], quoted=False))
                # set the catalog name
                if len(expanded_path) == 3:
                    if node.db != expanded_path[0]:
                        node.set("catalog", sqlglot.to_identifier(expanded_path[0], quoted=False))
        # quote and case-fold identifiers, TODO: maybe we could be more intelligent, but then we need to unquote ibis
        if isinstance(node, sge.Identifier):
            if is_casefolding:
                node.set("this", casefold_identifier(node.this))
            node.set("quoted", True)

    # add aliases to output selects to stay compatible with dlt schema after the query
    if orig_selects:
        for i, orig in orig_selects.items():
            case_folded_orig = casefold_identifier(orig)
            if case_folded_orig != orig:
                # somehow we need to alias just top select in UNION (tested on Snowflake)
                sel_expr = qualified_query.selects[i]
                qualified_query.selects[i] = sge.alias_(sel_expr, orig, quoted=True)

    return qualified_query



import inspect

from functools import wraps
from typing import Callable, Dict, Type, Any, Optional, Union, Tuple, TypeVar, overload, cast
from inspect import Signature, Parameter, unwrap

from dlt.common.typing import (
    DictStrAny,
    TFun,
    AnyFun,
    get_type_globals,
    resolve_single_annotation,
)
from dlt.common.configuration.resolve import resolve_configuration, inject_section
from dlt.common.configuration.specs.base_configuration import BaseConfiguration
from dlt.common.configuration.specs.config_section_context import ConfigSectionContext

from dlt.common.reflection.spec import spec_from_signature


_LAST_DLT_CONFIG = "_dlt_config"
_ORIGINAL_ARGS = "_dlt_orig_args"
_FUN_SPEC = "__SPEC__"
TConfiguration = TypeVar("TConfiguration", bound=BaseConfiguration)


def get_fun_spec(f: AnyFun) -> Type[BaseConfiguration]:
    return getattr(unwrap(f), _FUN_SPEC, None)  # type: ignore[no-any-return]


def get_fun_last_config(f: AnyFun) -> BaseConfiguration:
    return getattr(unwrap(f), _LAST_DLT_CONFIG, None)  # type: ignore[no-any-return]


def set_fun_spec(f: AnyFun, spec: Type[BaseConfiguration]) -> None:
    """Assigns a spec to a unwrapped `f` from which it was inferred"""
    # always add SPEC to unwrapped functions, skip all decorated functions
    setattr(unwrap(f), _FUN_SPEC, spec)  # noqa: B010


def set_fun_last_config(f: AnyFun, config: BaseConfiguration) -> None:
    """Assigns a resolved config to a unwrapped `f` when it is called"""
    setattr(unwrap(f), _LAST_DLT_CONFIG, config)  # noqa: B010


@overload
def with_config(
    func: TFun,
    /,
    spec: Type[BaseConfiguration] = None,
    sections: Union[str, Tuple[str, ...]] = (),
    section_arg_name: str = None,
    sections_merge_style: ConfigSectionContext.TMergeFunc = ConfigSectionContext.prefer_incoming,
    include_defaults: bool = True,
    accept_partial: bool = False,
    initial_config: BaseConfiguration = None,
    base: Type[BaseConfiguration] = BaseConfiguration,
    lock_context_on_injection: bool = True,
) -> TFun: ...


@overload
def with_config(
    func: None = ...,
    /,
    spec: Type[BaseConfiguration] = None,
    sections: Union[str, Tuple[str, ...]] = (),
    section_arg_name: str = None,
    sections_merge_style: ConfigSectionContext.TMergeFunc = ConfigSectionContext.prefer_incoming,
    include_defaults: bool = True,
    accept_partial: bool = False,
    initial_config: Optional[BaseConfiguration] = None,
    base: Type[BaseConfiguration] = BaseConfiguration,
    lock_context_on_injection: bool = True,
) -> Callable[[TFun], TFun]: ...


def with_config(
    func: Optional[AnyFun] = None,
    /,
    spec: Type[BaseConfiguration] = None,
    sections: Union[str, Tuple[str, ...]] = (),
    section_arg_name: str = None,
    sections_merge_style: ConfigSectionContext.TMergeFunc = ConfigSectionContext.prefer_incoming,
    include_defaults: bool = True,
    accept_partial: bool = False,
    initial_config: Optional[BaseConfiguration] = None,
    base: Type[BaseConfiguration] = BaseConfiguration,
    lock_context_on_injection: bool = True,
) -> Callable[[TFun], TFun]:
    """Injects values into decorated function arguments following the specification in `spec` or by deriving one from function's signature.

    The synthesized spec contains the arguments marked with `dlt.secrets.value` and `dlt.config.value` which are required to be injected at runtime.
    Optionally (and by default) arguments with default values are included in spec as well.

    Args:
        func (Optional[AnyFun], optional): A function with arguments to be injected. Defaults to None.
        spec (Type[BaseConfiguration], optional): A specification of injectable arguments. Defaults to None.
        sections (Tuple[str, ...], optional): A set of config sections in which to look for arguments values. Defaults to ().
        section_arg_name (bool, optional): Name of the argument in the signature of the decorated function which will be used to extend `sections` tuple.
            A top level pipeline section will be added if argument name is `pipeline_name`
        prefer_existing_sections: (bool, optional): When joining existing section context, the existing context will be preferred to the one in `sections`. Default: False
        include_defaults (bool, optional): If True then arguments with default values will be included in synthesized spec. If False only the required arguments marked with `dlt.secrets.value` and `dlt.config.value` are included
        base (Type[BaseConfiguration], optional): A base class for synthesized spec. Defaults to BaseConfiguration.
        lock_context_on_injection (bool, optional): If True, the thread context will be locked during injection to prevent race conditions. Defaults to True.
    Returns:
        Callable[[TFun], TFun]: A decorated function
    """

    def decorator(f: TFun) -> TFun:
        SPEC: Type[BaseConfiguration] = None
        sig: Signature = inspect.signature(f)
        signature_fields: Dict[str, Any]
        # find variadic kwargs to which additional arguments and injection context can be injected
        kwargs_arg = next(
            (
                p
                for p in sig.parameters.values()
                if p.kind == Parameter.VAR_KEYWORD and p.name == "injection_kwargs"
            ),
            None,
        )
        if spec is None:
            SPEC, signature_fields = spec_from_signature(f, sig, include_defaults, base=base)
        else:
            SPEC = spec
            signature_fields = SPEC.get_resolvable_fields()

        # if no signature fields were added we will not wrap `f` for injection
        if len(signature_fields) == 0:
            # always register new function
            set_fun_spec(f, SPEC)
            return f

        spec_arg: Parameter = None
        section_name_arg: Parameter = None
        globalns = get_type_globals(f)

        for p in sig.parameters.values():
            # for all positional parameters that do not have default value, set default
            # if hasattr(SPEC, p.name) and p.default == Parameter.empty:
            #     p._default = None  # type: ignore
            if resolve_single_annotation(p.annotation, globalns=globalns) is SPEC:
                # if any argument has type SPEC then us it to take initial value
                spec_arg = p
            if p.name == section_arg_name:
                # add value of section_name_arg to
                section_name_arg = p
                section_name_arg_default = None if p.default == Parameter.empty else p.default

        def resolve_config(
            bound_args: inspect.BoundArguments, accept_partial_: bool
        ) -> BaseConfiguration:
            """Resolve arguments using the provided spec"""
            # bind parameters to signature
            # for calls containing resolved spec in the kwargs, we do not need to resolve again
            config: BaseConfiguration = None

            curr_sections: Union[str, Tuple[str, ...]] = None
            # section may be a function from function arguments to section
            if callable(sections):
                curr_sections = sections(bound_args.arguments)
            else:
                curr_sections = sections
            # sections may be a string
            if isinstance(curr_sections, str):
                curr_sections = (curr_sections,)
            # extend sections with section_name_arg
            if section_name_arg:
                section_extension = bound_args.arguments.get(
                    section_name_arg.name, section_name_arg_default
                )
                if section_extension:
                    curr_sections = (
                        curr_sections + (section_extension,)
                        if curr_sections
                        else (section_extension,)
                    )

            # if one of arguments is spec the use it as initial value
            if initial_config:
                config = initial_config
            elif spec_arg:
                config = bound_args.arguments.get(spec_arg.name, None)
            # resolve SPEC, also provide section_context with pipeline_name
            if section_name_arg and section_name_arg.name == "pipeline_name":
                curr_pipeline_name = bound_args.arguments.get(
                    section_name_arg.name, section_name_arg_default
                )
            else:
                curr_pipeline_name = None
            section_context = ConfigSectionContext(
                pipeline_name=curr_pipeline_name,
                sections=curr_sections,
                merge_style=sections_merge_style,
            )

            # this may be called from many threads so section_context is thread affine
            with inject_section(section_context, lock_context=lock_context_on_injection):
                # print(
                #     f"RESOLVE CONF in inject: {SPEC} {f.__name__}: {section_context.sections} vs"
                #     f" {sections} in {bound_args.arguments} and partial {accept_partial_}"
                # )
                return resolve_configuration(
                    config or SPEC(),
                    explicit_value=bound_args.arguments,
                    accept_partial=accept_partial_,
                )

        def update_bound_args(
            bound_args: inspect.BoundArguments, config: BaseConfiguration, args: Any, kwargs: Any
        ) -> None:
            # overwrite or add resolved params
            resolved_params = dict(config)
            # overwrite or add resolved params
            for p in sig.parameters.values():
                if p.name in resolved_params:
                    bound_args.arguments[p.name] = resolved_params.pop(p.name)
                if resolve_single_annotation(p.annotation, globalns=globalns) is SPEC:
                    bound_args.arguments[p.name] = config
            # pass all other config parameters into kwargs if present
            if kwargs_arg is not None:
                if kwargs_arg.name not in bound_args.arguments:
                    # add variadic keyword argument
                    bound_args.arguments[kwargs_arg.name] = {}
                bound_args.arguments[kwargs_arg.name].update(resolved_params)
                bound_args.arguments[kwargs_arg.name][_LAST_DLT_CONFIG] = config
                bound_args.arguments[kwargs_arg.name][_ORIGINAL_ARGS] = (args, kwargs)

        def with_partially_resolved_config(config: Optional[BaseConfiguration] = None) -> Any:
            # creates a pre-resolved partial of the decorated function
            if not config:
                # TODO: this will not work if correct config is not provided
                # esp. in case of parameters in _wrap being ConfigurationBase
                # at least we should implement re-resolve with explicit parameters
                # so we can merge partial we get here to combine a full config
                empty_bound_args = sig.bind_partial()
                # TODO: resolve partial here that will be updated in _wrap
                config = resolve_config(empty_bound_args, accept_partial_=False)

            @wraps(f)
            def _wrap(*args: Any, **kwargs: Any) -> Any:
                # TODO: we should not change the outer config but deepcopy it
                nonlocal config

                # Do we need an exception here?
                if spec_arg and spec_arg.name in kwargs:
                    from dlt.common import logger

                    logger.warning(
                        "Spec argument is provided in kwargs, ignoring it for resolved partial"
                        " function."
                    )

                # we can still overwrite the config
                if _LAST_DLT_CONFIG in kwargs:
                    config = last_config(**kwargs)

                # call the function with the pre-resolved config
                bound_args = sig.bind(*args, **kwargs)
                # TODO: update partial config with bound_args (to cover edge cases with embedded configs)
                update_bound_args(bound_args, config, args, kwargs)
                set_fun_last_config(f, config)
                return f(*bound_args.args, **bound_args.kwargs)

            return _wrap

        @wraps(f)
        def _wrap(*args: Any, **kwargs: Any) -> Any:
            # Resolve config
            config: BaseConfiguration = None
            bound_args = sig.bind_partial(*args, **kwargs)
            if _LAST_DLT_CONFIG in kwargs:
                config = last_config(**kwargs)
            else:
                config = resolve_config(bound_args, accept_partial_=accept_partial)

            # call the function with resolved config
            update_bound_args(bound_args, config, args, kwargs)
            set_fun_last_config(f, config)
            return f(*bound_args.args, **bound_args.kwargs)

        set_fun_spec(f, SPEC)

        # add a method to create a pre-resolved partial
        setattr(_wrap, "__RESOLVED_PARTIAL_FUNC__", with_partially_resolved_config)  # noqa: B010

        return _wrap  # type: ignore

    # See if we're being called as @with_config or @with_config().
    if func is None:
        # We're called with parens.
        return decorator

    if not callable(func):
        raise ValueError(
            "First parameter of `with_config` must be callable i.e., by using it as a decorator"
            " `@with_config`"
        )

    # We're called as @with_config without parens.
    return decorator(func)


def last_config(**injection_kwargs: Any) -> Any:
    """Get configuration instance used to inject function kwargs"""
    return injection_kwargs[_LAST_DLT_CONFIG]


def get_orig_args(**injection_kwargs: Any) -> Tuple[Tuple[Any], DictStrAny]:
    """Get original argument with which the injectable function was called"""
    return injection_kwargs[_ORIGINAL_ARGS]  # type: ignore


def create_resolved_partial(f: AnyFun, config: Optional[BaseConfiguration] = None) -> AnyFun:
    """Create a pre-resolved partial of the with_config decorated function"""
    if partial_func := getattr(f, "__RESOLVED_PARTIAL_FUNC__", None):
        return cast(AnyFun, partial_func(config))
    return f



from typing import List, Dict, Set
from abc import abstractmethod

from dlt.common.schema import TSchemaUpdate, Schema
from dlt.common.storages.load_storage import LoadStorage
from dlt.common.storages import NormalizeStorage
from dlt.common.storages.data_item_storage import DataItemStorage

from dlt.normalize.configuration import NormalizeConfiguration


class ItemsNormalizer:
    def __init__(
        self,
        item_storage: DataItemStorage,
        load_storage: LoadStorage,
        normalize_storage: NormalizeStorage,
        schema: Schema,
        load_id: str,
        config: NormalizeConfiguration,
    ) -> None:
        self.item_storage = item_storage
        self.load_storage = load_storage
        self.normalize_storage = normalize_storage
        self.schema = schema
        self.load_id = load_id
        self.config = config
        self.naming = self.schema.naming

    def _maybe_cancel(self) -> None:
        self.load_storage.new_packages.raise_if_cancelled(self.load_id)

    @property
    def null_only_columns(self) -> Dict[str, Set[str]]:
        return {}

    @abstractmethod
    def __call__(self, extracted_items_file: str, root_table_name: str) -> List[TSchemaUpdate]: ...

    def close(self) -> None:
        """Called when normalizer finishes processing. Override to clean up resources."""
        pass



import os
from typing import List

from dlt.common import known_env
from dlt.common.utils import uniq_id_base64, many_uniq_ids_base64


DLT_ID_LENGTH_BYTES = int(os.environ.get(known_env.DLT_DLT_ID_LENGTH_BYTES, 10))


def generate_dlt_ids(n_ids: int) -> List[str]:
    return many_uniq_ids_base64(n_ids, DLT_ID_LENGTH_BYTES)


def generate_dlt_id() -> str:
    return uniq_id_base64(DLT_ID_LENGTH_BYTES)



"""
Ports fragments of URL class from Sql Alchemy to use them when dependency is not available.
"""

from typing import cast

# port basic functionality without the whole Sql Alchemy

import re
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)
import collections.abc as collections_abc
from urllib.parse import (
    quote_plus,
    parse_qsl,
    quote,
    unquote,
)
from dlt.common.exceptions import TypeErrorWithKnownTypes

_KT = TypeVar("_KT", bound=Any)
_VT = TypeVar("_VT", bound=Any)


class ImmutableDict(Dict[_KT, _VT]):
    """Not a real immutable dict"""

    def __setitem__(self, __key: _KT, __value: _VT) -> None:
        raise NotImplementedError("Cannot modify immutable dict")

    def __delitem__(self, _KT: Any) -> None:
        raise NotImplementedError("Cannot modify immutable dict")

    def update(self, *arg: Any, **kw: Any) -> None:
        raise NotImplementedError("Cannot modify immutable dict")


EMPTY_DICT: ImmutableDict[Any, Any] = ImmutableDict()


def to_list(value: Any, default: Optional[List[Any]] = None) -> List[Any]:
    if value is None:
        return default
    if not isinstance(value, collections_abc.Iterable) or isinstance(value, str):
        return [value]
    elif isinstance(value, list):
        return value
    else:
        return list(value)


class URL(NamedTuple):
    """
    Represent the components of a URL used to connect to a database.

    Based on SqlAlchemy URL class with copyright as below:

    # engine/url.py
    # Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
    #
    # This module is part of SQLAlchemy and is released under
    # the MIT License: https://www.opensource.org/licenses/mit-license.php
    """

    drivername: str
    """database backend and driver name, such as `postgresql+psycopg2`"""
    username: Optional[str]
    "username string"
    password: Optional[str]
    """password, which is normally a string but may also be any object that has a `__str__()` method."""
    host: Optional[str]
    """hostname or IP number.  May also be a data source name for some drivers."""
    port: Optional[int]
    """integer port number"""
    database: Optional[str]
    """database name"""
    query: ImmutableDict[str, Union[Tuple[str, ...], str]]
    """an immutable mapping representing the query string.  contains strings
    for keys and either strings or tuples of strings for values"""

    @classmethod
    def create(
        cls,
        drivername: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        database: Optional[str] = None,
        query: Mapping[str, Union[Sequence[str], str]] = None,
    ) -> "URL":
        """Create a new `URL` object."""
        return cls(
            cls._assert_str(drivername, "drivername"),
            cls._assert_none_str(username, "username"),
            password,
            cls._assert_none_str(host, "host"),
            cls._assert_port(port),
            cls._assert_none_str(database, "database"),
            cls._str_dict(query or EMPTY_DICT),
        )

    @classmethod
    def _assert_port(cls, port: Optional[int]) -> Optional[int]:
        if port is None:
            return None
        try:
            return int(port)
        except TypeError:
            raise TypeErrorWithKnownTypes("port", port, ["int", None])

    @classmethod
    def _assert_str(cls, v: str, paramname: str) -> str:
        if not isinstance(v, str):
            raise TypeErrorWithKnownTypes("paramname", paramname, ["str"])
        return v

    @classmethod
    def _assert_none_str(cls, v: Optional[str], paramname: str) -> Optional[str]:
        if v is None:
            return v

        return cls._assert_str(v, paramname)

    @classmethod
    def _str_dict(
        cls,
        dict_: Optional[
            Union[
                Sequence[Tuple[str, Union[Sequence[str], str]]],
                Mapping[str, Union[Sequence[str], str]],
            ]
        ],
    ) -> ImmutableDict[str, Union[Tuple[str, ...], str]]:
        if dict_ is None:
            return EMPTY_DICT

        @overload
        def _assert_value(
            val: str,
        ) -> str: ...

        @overload
        def _assert_value(
            val: Sequence[str],
        ) -> Union[str, Tuple[str, ...]]: ...

        def _assert_value(
            val: Union[str, Sequence[str]],
        ) -> Union[str, Tuple[str, ...]]:
            if isinstance(val, str):
                return val
            elif isinstance(val, collections_abc.Sequence):
                return tuple(_assert_value(elem) for elem in val)
            else:
                raise TypeErrorWithKnownTypes("value", val, ["str", "Sequence[str]"])

        def _assert_str(v: str) -> str:
            if not isinstance(v, str):
                raise TypeErrorWithKnownTypes("key", v, ["str"])
            return v

        dict_items: Iterable[Tuple[str, Union[Sequence[str], str]]]
        if isinstance(dict_, collections_abc.Sequence):
            dict_items = dict_
        else:
            dict_items = dict_.items()

        return ImmutableDict(
            {
                _assert_str(key): _assert_value(
                    value,
                )
                for key, value in dict_items
            }
        )

    def set(  # noqa
        self,
        drivername: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        database: Optional[str] = None,
        query: Optional[Mapping[str, Union[Sequence[str], str]]] = None,
    ) -> "URL":
        """return a new `URL` object with modifications."""

        kw: Dict[str, Any] = {}
        if drivername is not None:
            kw["drivername"] = drivername
        if username is not None:
            kw["username"] = username
        if password is not None:
            kw["password"] = password
        if host is not None:
            kw["host"] = host
        if port is not None:
            kw["port"] = port
        if database is not None:
            kw["database"] = database
        if query is not None:
            kw["query"] = query

        return self._assert_replace(**kw)

    def _assert_replace(self, **kw: Any) -> "URL":
        """argument checks before calling _replace()"""

        if "drivername" in kw:
            self._assert_str(kw["drivername"], "drivername")
        for name in "username", "host", "database":
            if name in kw:
                self._assert_none_str(kw[name], name)
        if "port" in kw:
            self._assert_port(kw["port"])
        if "query" in kw:
            kw["query"] = self._str_dict(kw["query"])

        return self._replace(**kw)

    def update_query_string(self, query_string: str, append: bool = False) -> "URL":
        return self.update_query_pairs(parse_qsl(query_string), append=append)

    def update_query_pairs(
        self,
        key_value_pairs: Iterable[Tuple[str, Union[str, List[str]]]],
        append: bool = False,
    ) -> "URL":
        """Return a new `URL` object with the `query` parameter dictionary updated by the given sequence of key/value pairs"""
        existing_query = self.query
        new_keys: Dict[str, Union[str, List[str]]] = {}

        for key, value in key_value_pairs:
            if key in new_keys:
                new_keys[key] = to_list(new_keys[key])
                cast("List[str]", new_keys[key]).append(cast(str, value))
            else:
                new_keys[key] = to_list(value) if isinstance(value, (list, tuple)) else value

        new_query: Mapping[str, Union[str, Sequence[str]]]
        if append:
            new_query = {}

            for k in new_keys:
                if k in existing_query:
                    new_query[k] = tuple(to_list(existing_query[k]) + to_list(new_keys[k]))
                else:
                    new_query[k] = new_keys[k]

            new_query.update(
                {k: existing_query[k] for k in set(existing_query).difference(new_keys)}
            )
        else:
            new_query = ImmutableDict(
                {
                    **self.query,
                    **{k: tuple(v) if isinstance(v, list) else v for k, v in new_keys.items()},
                }
            )
        return self.set(query=new_query)

    def update_query_dict(
        self,
        query_parameters: Mapping[str, Union[str, List[str]]],
        append: bool = False,
    ) -> "URL":
        return self.update_query_pairs(query_parameters.items(), append=append)

    def render_as_string(self, hide_password: bool = True) -> str:
        """Render this `URL` object as a string."""
        s = self.drivername + "://"
        if self.username is not None:
            s += quote(self.username, safe=" +")
            if self.password is not None:
                s += ":" + ("***" if hide_password else quote(str(self.password), safe=" +"))
            s += "@"
        if self.host is not None:
            if ":" in self.host:
                s += f"[{self.host}]"
            else:
                s += self.host
        if self.port is not None:
            s += ":" + str(self.port)
        if self.database is not None:
            s += "/" + self.database
        if self.query:
            keys = to_list(self.query)
            keys.sort()
            s += "?" + "&".join(
                f"{quote_plus(k)}={quote_plus(element)}"
                for k in keys
                for element in to_list(self.query[k])
            )
        return s

    def __repr__(self) -> str:
        return self.render_as_string()

    def __copy__(self) -> "URL":
        return self.__class__.create(
            self.drivername,
            self.username,
            self.password,
            self.host,
            self.port,
            self.database,
            self.query.copy(),
        )

    def __deepcopy__(self, memo: Any) -> "URL":
        return self.__copy__()

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, URL)
            and self.drivername == other.drivername
            and self.username == other.username
            and self.password == other.password
            and self.host == other.host
            and self.database == other.database
            and self.query == other.query
            and self.port == other.port
        )

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def get_backend_name(self) -> str:
        """Return the backend name.

        This is the name that corresponds to the database backend in
        use, and is the portion of the `drivername`
        that is to the left of the plus sign.

        """
        if "+" not in self.drivername:
            return self.drivername
        else:
            return self.drivername.split("+")[0]

    def get_driver_name(self) -> str:
        """Return the backend name.

        This is the name that corresponds to the DBAPI driver in
        use, and is the portion of the `drivername`
        that is to the right of the plus sign.
        """

        if "+" not in self.drivername:
            return self.drivername
        else:
            return self.drivername.split("+")[1]


def make_url(name_or_url: Union[str, URL]) -> URL:
    """Given a string, produce a new URL instance.

    The format of the URL generally follows `RFC-1738`, with some exceptions, including
    that underscores, and not dashes or periods, are accepted within the
    "scheme" portion.

    If a `URL` object is passed, it is returned as is."""

    if isinstance(name_or_url, str):
        return _parse_url(name_or_url)
    elif not isinstance(name_or_url, URL):
        raise ValueError(f"Expected string or URL object, got `{name_or_url!r}`")
    else:
        return name_or_url


def _parse_url(name: str) -> URL:
    pattern = re.compile(
        r"""
            (?P<name>[\w\+]+)://
            (?:
                (?P<username>[^:/]*)
                (?::(?P<password>[^@]*))?
            @)?
            (?:
                (?:
                    \[(?P<ipv6host>[^/\?]+)\] |
                    (?P<ipv4host>[^/:\?]+)
                )?
                (?::(?P<port>[^/\?]*))?
            )?
            (?:/(?P<database>[^\?]*))?
            (?:\?(?P<query>.*))?
            """,
        re.X,
    )

    m = pattern.match(name)
    if m is not None:
        components = m.groupdict()
        query: Optional[Dict[str, Union[str, List[str]]]]
        if components["query"] is not None:
            query = {}

            for key, value in parse_qsl(components["query"]):
                if key in query:
                    query[key] = to_list(query[key])
                    cast("List[str]", query[key]).append(value)
                else:
                    query[key] = value
        else:
            query = None

        components["query"] = query
        if components["username"] is not None:
            components["username"] = unquote(components["username"])

        if components["password"] is not None:
            components["password"] = unquote(components["password"])

        ipv4host = components.pop("ipv4host")
        ipv6host = components.pop("ipv6host")
        components["host"] = ipv4host or ipv6host
        name = components.pop("name")

        if components["port"]:
            components["port"] = int(components["port"])

        return URL.create(name, **components)  # type: ignore

    else:
        raise ValueError(f"Could not parse SQLAlchemy URL from string '{name}'")



import io
import glob
import gzip
import mimetypes
import pathlib
import posixpath
from io import BytesIO
from typing import (
    Literal,
    cast,
    Tuple,
    Optional,
    Union,
    Iterator,
    Any,
    IO,
    Dict,
    Callable,
    Sequence,
)
from urllib.parse import urlparse
from typing_extensions import NotRequired

from fsspec import AbstractFileSystem, register_implementation, get_filesystem_class
from fsspec.core import url_to_fs

from dlt import version
from dlt.common.typing import TypedDict
from dlt.common.pendulum import pendulum
from dlt.common.configuration.specs import (
    GcpCredentials,
    AwsCredentials,
    AzureCredentials,
    HfCredentials,
    SFTPCredentials,
)
from dlt.common.exceptions import MissingDependencyException, ValueErrorWithKnownValues
from dlt.common.storages.configuration import (
    FileSystemCredentials,
    FilesystemConfiguration,
    make_fsspec_url,
)
from dlt.common.time import ensure_pendulum_datetime_utc
from dlt.common.typing import DictStrAny
from dlt.common.utils import without_none


class FileItem(TypedDict):
    """A DataItem representing a file"""

    file_url: str
    file_name: str
    relative_path: str
    mime_type: str
    encoding: NotRequired[str]
    modification_date: pendulum.DateTime
    size_in_bytes: int
    file_content: NotRequired[bytes]


# Map of protocol to mtime resolver
# we only need to support a small finite set of protocols
MTIME_DISPATCH = {
    "s3": lambda f: ensure_pendulum_datetime_utc(f["LastModified"]),
    "adl": lambda f: ensure_pendulum_datetime_utc(f["LastModified"]),
    "az": lambda f: ensure_pendulum_datetime_utc(f["last_modified"]),
    "gcs": lambda f: ensure_pendulum_datetime_utc(f["updated"]),
    "hf": lambda f: ensure_pendulum_datetime_utc(f["last_commit"]["date"]),
    "https": lambda f: cast(
        pendulum.DateTime,
        pendulum.parse(
            f.get("Last-Modified", pendulum.now().isoformat()), exact=True, strict=False
        ),
    ),
    "http": lambda f: cast(
        pendulum.DateTime,
        pendulum.parse(
            f.get("Last-Modified", pendulum.now().isoformat()), exact=True, strict=False
        ),
    ),
    "file": lambda f: ensure_pendulum_datetime_utc(f["mtime"]),
    "memory": lambda f: ensure_pendulum_datetime_utc(f["created"]),
    "gdrive": lambda f: ensure_pendulum_datetime_utc(f["modifiedTime"]),
    "sftp": lambda f: ensure_pendulum_datetime_utc(f["mtime"]),
}
# Support aliases
MTIME_DISPATCH["gs"] = MTIME_DISPATCH["gcs"]
MTIME_DISPATCH["s3a"] = MTIME_DISPATCH["s3"]
MTIME_DISPATCH["abfs"] = MTIME_DISPATCH["az"]
MTIME_DISPATCH["abfss"] = MTIME_DISPATCH["az"]

# Map of protocol to a filesystem type
CREDENTIALS_DISPATCH: Dict[str, Callable[[FilesystemConfiguration], DictStrAny]] = {
    "s3": lambda config: cast(AwsCredentials, config.credentials).to_s3fs_credentials(),
    "az": lambda config: cast(AzureCredentials, config.credentials).to_adlfs_credentials(),
    "gs": lambda config: cast(GcpCredentials, config.credentials).to_gcs_credentials(),
    "hf": lambda config: cast(HfCredentials, config.credentials).to_hffs_credentials(),
    "gdrive": lambda config: {"credentials": cast(GcpCredentials, config.credentials)},
    "sftp": lambda config: cast(SFTPCredentials, config.credentials).to_fsspec_credentials(),
}
CREDENTIALS_DISPATCH["adl"] = CREDENTIALS_DISPATCH["az"]
CREDENTIALS_DISPATCH["abfs"] = CREDENTIALS_DISPATCH["az"]
CREDENTIALS_DISPATCH["azure"] = CREDENTIALS_DISPATCH["az"]
CREDENTIALS_DISPATCH["abfss"] = CREDENTIALS_DISPATCH["az"]
CREDENTIALS_DISPATCH["gcs"] = CREDENTIALS_DISPATCH["gs"]

# Default kwargs for protocol
DEFAULT_KWARGS: Dict[str, Dict[str, Any]] = {
    # disable concurrent
    "az": {"max_concurrency": 1},
    # get last_commit info which includes last_commit.date
    "hf": {"expand_info": True},
}
DEFAULT_KWARGS["adl"] = DEFAULT_KWARGS["az"]
DEFAULT_KWARGS["abfs"] = DEFAULT_KWARGS["az"]
DEFAULT_KWARGS["azure"] = DEFAULT_KWARGS["az"]
DEFAULT_KWARGS["abfss"] = DEFAULT_KWARGS["az"]

AZURE_BLOB_STORAGE_PROTOCOLS = ["az", "azure", "adl", "abfss", "abfs"]
S3_PROTOCOLS = ["s3", "s3a"]
GCS_PROTOCOLS = ["gs", "gcs"]


def fsspec_filesystem(
    protocol: str,
    credentials: FileSystemCredentials = None,
    kwargs: Optional[DictStrAny] = None,
    client_kwargs: Optional[DictStrAny] = None,
    config_kwargs: Optional[DictStrAny] = None,
) -> Tuple[AbstractFileSystem, str]:
    """Instantiates an authenticated fsspec `FileSystem` for a given `protocol` and credentials.

    Please supply credentials instance corresponding to the protocol.
    The `protocol` is just the code name of the filesystem i.e.:
    * s3
    * az, abfs, abfss, adl, azure
    * gcs, gs

    also see filesystem_from_config
    """
    return fsspec_from_config(
        FilesystemConfiguration(
            protocol,
            credentials,
            kwargs=kwargs,
            client_kwargs=client_kwargs,
            config_kwargs=config_kwargs,
        )
    )


def prepare_fsspec_args(config: FilesystemConfiguration) -> DictStrAny:
    """Prepare arguments for fsspec filesystem constructor.

    Args:
        config (FilesystemConfiguration): The filesystem configuration.

    Returns:
        DictStrAny: The arguments for the fsspec filesystem constructor.
    """
    protocol = config.protocol
    # never use listing caches
    fs_kwargs: DictStrAny = {
        "use_listings_cache": config.read_only,
        "listings_expiry_time": 60.0,
        "skip_instance_cache": False,
    }
    credentials = CREDENTIALS_DISPATCH.get(protocol, lambda _: {})(config)

    if protocol == "gdrive":
        from dlt.common.storages.fsspecs.google_drive import GoogleDriveFileSystem

        register_implementation("gdrive", GoogleDriveFileSystem, "GoogleDriveFileSystem")

    fs_kwargs.update(DEFAULT_KWARGS.get(protocol, {}))

    if protocol in ("https", "http", "sftp"):
        fs_kwargs.clear()

    if config.kwargs is not None:
        fs_kwargs.update(config.kwargs)
    if config.client_kwargs is not None:
        fs_kwargs["client_kwargs"] = config.client_kwargs

    if "client_kwargs" in fs_kwargs and "client_kwargs" in credentials:
        fs_kwargs["client_kwargs"].update(credentials.pop("client_kwargs"))

    if "config_kwargs" in fs_kwargs and "config_kwargs" in credentials:
        fs_kwargs["config_kwargs"].update(credentials.pop("config_kwargs"))

    fs_kwargs.update(without_none(credentials))
    return fs_kwargs


def fsspec_from_config(config: FilesystemConfiguration) -> Tuple[AbstractFileSystem, str]:
    """Instantiates an authenticated fsspec `FileSystem` from `config` argument.

    Authenticates following filesystems:
    * s3
    * az, abfs, abfss, adl, azure
    * gcs, gs
    * sftp
    * hf

    All other filesystems are not authenticated

    Returns: (fsspec filesystem, normalized url)
    """
    fs_kwargs = prepare_fsspec_args(config)

    try:
        # first get the class to check the protocol
        fs_cls = get_filesystem_class(config.protocol)
        if fs_cls.protocol == "abfs":
            url = urlparse(config.bucket_url)
            # if storage account is present in bucket_url and in credentials, az fsspec will fail
            # account name is detected for blob.core.windows.net and dfs.core.windows.net hosts
            if url.username and (
                url.hostname.endswith("blob.core.windows.net")
                or url.hostname.endswith("dfs.core.windows.net")
            ):
                fs_kwargs.pop("account_name")
        return url_to_fs(config.bucket_url, **fs_kwargs)  # type: ignore
    except ImportError as e:
        raise MissingDependencyException(
            "filesystem", [f"{version.DLT_PKG_NAME}[{config.protocol}]"]
        ) from e


class FileItemDict(DictStrAny):
    """A FileItem dictionary with additional methods to get fsspec filesystem, open and read files."""

    def __init__(
        self,
        mapping: FileItem,
        fsspec: AbstractFileSystem = None,
    ):
        """Create a dictionary with the filesystem client.

        Args:
            mapping (FileItem): The file item TypedDict.
            credentials (Optional[FileSystemCredentials], optional): The credentials to the
                filesystem. Defaults to None.
        """
        self._fsspec = fsspec
        super().__init__(**mapping)

    @property
    def fsspec(self) -> AbstractFileSystem:
        """The filesystem client is based on the given credentials.

        Returns:
            AbstractFileSystem: The fsspec client.
        """
        return self._fsspec

    @property
    def local_file_path(self) -> str:
        """Gets a valid local filesystem path from file:// scheme.
        Supports POSIX/Windows/UNC paths

        Returns:
            str: local filesystem path
        """
        return FilesystemConfiguration.make_local_path(self["file_url"])

    def open(  # noqa: A003
        self,
        mode: str = "rb",
        compression: Literal["auto", "disable", "enable"] = "auto",
        **kwargs: Any,
    ) -> IO[Any]:
        """Open the file as a fsspec file.

        This method opens the file represented by this dictionary as a file-like object using
        the fsspec library.

        Args:
            mode (Optional[str]): Open mode.
            compression (Optional[str]): A flag to enable/disable compression.
                Can have one of three values: "disable" - no compression applied,
                "enable" - gzip compression applied, "auto" (default) -
                compression applied only for files compressed with gzip.
            **kwargs (Any): The arguments to pass to the fsspec open function.

        Returns:
            IOBase: The fsspec file.
        """
        if compression == "auto":
            compression_arg = "gzip" if self.get("encoding") == "gzip" else None
        elif compression == "enable":
            compression_arg = "gzip"
        elif compression == "disable":
            compression_arg = None
        else:
            raise ValueErrorWithKnownValues(
                "compression", compression, ["auto", "enable", "disable"]
            )

        # if the user has already extracted the content, we use it so there is no need to
        # download the file again.
        if "file_content" in self:
            content = (
                gzip.decompress(self["file_content"])
                if compression_arg == "gzip"
                else self["file_content"]
            )
            bytes_io = BytesIO(content)

            if "t" not in mode:
                return bytes_io
            text_kwargs = {
                k: kwargs.pop(k) for k in ["encoding", "errors", "newline"] if k in kwargs
            }
            return io.TextIOWrapper(
                bytes_io,
                **text_kwargs,
            )
        # `FileItemDict` kwarg `fsspec` is `Optional`. If `fsspec=None` this code branch
        # will fail.
        else:
            if "file" in self.fsspec.protocol:
                # use native local file path to open file:// uris
                file_url = self.local_file_path
            else:
                file_url = self["file_url"]
            return self.fsspec.open(  # type: ignore[no-any-return]
                file_url, mode=mode, compression=compression_arg, **kwargs
            )

    def read_bytes(self) -> bytes:
        """Read the file content.

        Returns:
            bytes: The file content.
        """
        if "file_content" in self and self["file_content"] is not None:
            return self["file_content"]  # type: ignore
        else:
            with self.open(mode="rb", compression="disable") as f:
                return f.read()  # type: ignore[no-any-return]


def guess_mime_type(file_name: str) -> Sequence[str]:
    type_ = list(mimetypes.guess_type(posixpath.basename(file_name), strict=False))

    if not type_[0]:
        type_[0] = "application/" + (posixpath.splitext(file_name)[1][1:] or "octet-stream")

    return type_


def glob_files(
    fs_client: AbstractFileSystem, bucket_url: str, file_glob: str = "**"
) -> Iterator[FileItem]:
    """Get the files from the filesystem client.

    Args:
        fs_client (AbstractFileSystem): The filesystem client.
        bucket_url (str): The url to the bucket.
        file_glob (str): A glob for the filename filter.

    Returns:
        Iterable[FileItem]: The list of files.
    """
    is_local_fs = "file" in fs_client.protocol
    if is_local_fs and FilesystemConfiguration.is_local_path(bucket_url):
        bucket_url = FilesystemConfiguration.make_file_url(bucket_url)
    bucket_url_parsed = urlparse(bucket_url)

    if is_local_fs:
        root_dir = FilesystemConfiguration.make_local_path(bucket_url)
        # use a Python glob to get files
        files = glob.glob(str(pathlib.Path(root_dir).joinpath(file_glob)), recursive=True)
        glob_result = {file: fs_client.info(file) for file in files}
    else:
        # convert to fs_path
        root_dir = fs_client._strip_protocol(bucket_url)
        filter_url = posixpath.join(root_dir, file_glob)
        # for `hf` we need to invalidate the cache to ensure fresh data (not sure why, maybe their
        # fsspec impl does not invalidate cache like other fsspec impls)
        if fs_client.protocol == "hf":
            fs_client.invalidate_cache(filter_url)
        glob_result = fs_client.glob(filter_url, detail=True)
        if isinstance(glob_result, list):
            raise NotImplementedError(
                "Cannot request details when using fsspec.glob. For adlfs (Azure) please use"
                " version 2023.9.0 or later"
            )

    for file, md in glob_result.items():
        if md["type"] != "file":
            continue
        scheme = bucket_url_parsed.scheme

        # relative paths are always POSIX
        if is_local_fs:
            # use OS pathlib for local paths
            loc_path = pathlib.Path(file)
            file_name = loc_path.name
            rel_path = loc_path.relative_to(root_dir).as_posix()
            file_url = FilesystemConfiguration.make_file_url(file)
        else:
            file_name = posixpath.basename(file)
            rel_path = posixpath.relpath(file, root_dir)
            file_url = make_fsspec_url(scheme, file, bucket_url)

        mime_type, encoding = guess_mime_type(rel_path)
        file_item = FileItem(
            file_name=file_name,
            relative_path=rel_path,
            file_url=file_url,
            mime_type=mime_type,
            modification_date=MTIME_DISPATCH[scheme](md),
            size_in_bytes=int(md["size"]),
        )
        if encoding is not None:
            file_item["encoding"] = encoding
        yield file_item



from typing import Any, Dict

from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.typing import TColumnNames
from dlt.destinations.utils import get_resource_for_adapter
from dlt.extract import DltResource
from dlt.extract.items import TTableHintTemplate


VECTORIZE_HINT = "x-lancedb-embed"
NO_REMOVE_ORPHANS_HINT = "x-lancedb-remove-orphans"


def lancedb_adapter(
    data: Any,
    embed: TColumnNames = None,
    merge_key: TColumnNames = None,
    no_remove_orphans: bool = False,
) -> DltResource:
    """Prepares data for the LanceDB destination by specifying which columns should be embedded.

    Args:
        data (Any): The data to be transformed. It can be raw data or an instance
            of DltResource. If raw data, the function wraps it into a DltResource
            object.
        embed (TColumnNames, optional): Specify columns to generate embeddings for.
            It can be a single column name as a string, or a list of column names.
        merge_key (TColumnNames, optional): Specify columns to merge on.
            It can be a single column name as a string, or a list of column names.
        no_remove_orphans (bool): Specify whether to remove orphaned records in child
            tables with no parent records after merges to maintain referential integrity.

    Returns:
        DltResource: A resource with applied LanceDB-specific hints.

    Raises:
        ValueError: If input for `embed` invalid or empty.

    Examples:
        >>> data = [{"name": "Marcel", "description": "Moonbase Engineer"}]
        >>> lancedb_adapter(data, embed="description")
        [DltResource with hints applied]
    """
    resource = get_resource_for_adapter(data)

    additional_table_hints: Dict[str, TTableHintTemplate[Any]] = {}
    column_hints: TTableSchemaColumns = None

    if embed:
        if isinstance(embed, str):
            embed = [embed]
        if not isinstance(embed, list):
            raise ValueError(
                "`embed` must be a list of column names or a single column name as a string."
            )
        column_hints = {}

        # TODO: warn if hint exists and we override nullable
        for column_name in embed:
            column_hints[column_name] = {
                "name": column_name,
                VECTORIZE_HINT: True,  # type: ignore[misc]
                "nullable": True,  # must be nullable because lance will override it anyway
            }

    additional_table_hints[NO_REMOVE_ORPHANS_HINT] = no_remove_orphans

    if column_hints or additional_table_hints or merge_key:
        resource.apply_hints(
            merge_key=merge_key, columns=column_hints, additional_table_hints=additional_table_hints
        )
    else:
        raise ValueError(
            "You must must provide at least either the `embed` or `merge_key` or `remove_orphans`"
            " argument if using the adapter."
        )

    return resource



from typing import Union, Iterable, Optional, List, Dict, Any, Tuple
from itertools import chain
from dataclasses import dataclass

from dlt.common.schema import Schema
from dlt.common.typing import TypedDict
from dlt.common.schema.typing import TSimpleRegex, TTableSchema
from dlt.common.schema.utils import (
    group_tables_by_resource,
    compile_simple_regexes,
    compile_simple_regex,
    has_table_seen_data,
)
from dlt.common import jsonpath
from dlt.common.typing import REPattern

from dlt.extract.state import (
    TPipelineState,
    _sources_state,
    _get_matching_resources,
    _get_matching_sources,
    reset_resource_state,
    delete_source_state_keys,
)


class _DropInfo(TypedDict):
    """Collects all entities that will be affected by pipeline drop"""

    tables: List[str]
    tables_with_data: List[str]
    resource_states: List[str]
    resource_names: List[str]
    state_paths: List[str]
    schema_name: str
    drop_all: bool
    resource_pattern: Optional[REPattern]
    warnings: List[str]


@dataclass
class _DropDryRunResult:
    """Collects all modifications that will be executed as part of the pipeline drop."""

    schema: Schema
    state: TPipelineState
    info: _DropInfo
    modified_tables: List[TTableSchema]


def _create_modified_state(
    state: TPipelineState,
    resource_pattern: Optional[REPattern],
    source_name: str,
    state_paths: jsonpath.TAnyJsonPath,
    info: _DropInfo,
) -> Tuple[TPipelineState, _DropInfo]:
    all_source_states = _sources_state(state)
    # note that source name is the same as its schema name and only resources belonging
    # to that source will be reset there
    source_state = all_source_states.get(source_name)
    if source_state:
        # drop table states
        if resource_pattern:
            for key in _get_matching_resources(resource_pattern, source_state):
                info["resource_states"].append(key)
                reset_resource_state(key, source_state)
        # drop additional state paths
        # Don't drop 'resources' key if jsonpath is wildcard
        resolved_paths = [
            p for p in jsonpath.resolve_paths(state_paths, source_state) if p != "resources"
        ]
        if state_paths and not resolved_paths:
            info["warnings"].append(
                f"State paths {state_paths} did not select any paths in source {source_name}"
            )
        delete_source_state_keys(resolved_paths, source_state)
        info["state_paths"].extend(f"{source_name}.{p}" for p in resolved_paths)
    return state, info


def prepare_drop_resources(
    schema: Schema,
    state: TPipelineState,
    resources: Union[Iterable[Union[str, TSimpleRegex]], Union[str, TSimpleRegex]] = (),
    state_paths: jsonpath.TAnyJsonPath = (),
    drop_all: bool = False,
    state_only: bool = False,
    # sources: Optional[Union[Iterable[Union[str, TSimpleRegex]], Union[str, TSimpleRegex]]] = None,
) -> _DropDryRunResult:
    """Generate a new schema and pipeline state with the requested resources removed.

    Args:
        schema: The schema to modify. Note that schema is changed in place.
        state: The pipeline state to modify. Note that state is changed in place.
        resources: Resource name(s) or regex pattern(s) matching resource names to drop.
            If empty, no resources will be dropped unless `drop_all` is True.
        state_paths: JSON path(s) relative to the source state to drop.
        drop_all: If True, all resources will be dropped (supersedes `resources`).
        state_only: If True, only modify the pipeline state, not schema
        sources: Only wipe state for sources matching the name(s) or regex pattern(s) in this list
            If not set all source states will be modified according to `state_paths` and `resources`

    Returns:
        A 3 part tuple containing the new schema, the new pipeline state, and a dictionary
        containing information about the drop operation.
    """

    if isinstance(resources, str):
        resources = [resources]
    resources = list(resources)
    if isinstance(state_paths, str):
        state_paths = [state_paths]

    state_paths = jsonpath.compile_paths(state_paths)

    resources = set(resources)
    if drop_all:
        resource_pattern = compile_simple_regex(TSimpleRegex("re:.*"))  # Match everything
    elif resources:
        resource_pattern = compile_simple_regexes(TSimpleRegex(r) for r in resources)
    else:
        resource_pattern = None

    if resource_pattern:
        # (1) Don't remove _dlt tables (2) Drop all selected tables from the schema
        # (3) Mark tables that seen data to be dropped in destination
        data_tables = {t["name"]: t for t in schema.data_tables(include_incomplete=True)}
        resource_tables = group_tables_by_resource(data_tables, pattern=resource_pattern)
        resource_names = list(resource_tables.keys())
        tables_to_drop_from_schema = list(chain.from_iterable(resource_tables.values()))
        tables_to_drop_from_schema.reverse()
        tables_to_drop_from_schema_names = [t["name"] for t in tables_to_drop_from_schema]
        tables_to_drop_from_dest = [t for t in tables_to_drop_from_schema if has_table_seen_data(t)]
    else:
        tables_to_drop_from_schema_names = []
        tables_to_drop_from_dest = []
        tables_to_drop_from_schema = []
        resource_names = []

    info: _DropInfo = dict(
        tables=tables_to_drop_from_schema_names if not state_only else [],
        tables_with_data=[t["name"] for t in tables_to_drop_from_dest] if not state_only else [],
        resource_states=[],
        state_paths=[],
        resource_names=resource_names if not state_only else [],
        schema_name=schema.name,
        drop_all=drop_all,
        resource_pattern=resource_pattern,
        warnings=[],
    )

    new_state, info = _create_modified_state(
        state, resource_pattern, schema.name, state_paths, info
    )
    # info["resource_names"] = resource_names if not state_only else []

    if resource_pattern and not resource_tables:
        info["warnings"].append(
            f"Specified resource(s) {str(resources)} did not select any table(s) in schema"
            f" {schema.name}. Possible resources are:"
            f" {list(group_tables_by_resource(data_tables).keys())}"
        )

    if not state_only:
        # drop only the selected tables
        schema.drop_tables(tables_to_drop_from_schema_names)
    return _DropDryRunResult(schema, new_state, info, tables_to_drop_from_dest)
